"""
Imported from the HoVer-Net repo with minor mods.

MIT License

Copyright (c) 2020 vqdang

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import numpy as np
from scipy.ndimage import distance_transform_edt

__all__ = ["gen_weight_maps"]


def gen_weight_maps(
    inst_map: np.ndarray, sigma: float = 5.0, w0: float = 10.0
) -> np.ndarray:
    """Generate a weight map like in U-Net paper.

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).
        sigma : float
            Sigma factor.
        w0 : float
            Weight multiplied to the penalty map.

    Returns
    -------
        np.ndarray:
            Object boundary boundary weight map. Shape (H, W).
    """
    inst_list = list(np.unique(inst_map))
    if 0 in inst_list:
        inst_list.remove(0)

    if len(inst_list) <= 1:  # 1 instance only
        return np.zeros(inst_map.shape[:2])

    stacked_inst_bgd_dst = np.zeros(inst_map.shape[:2] + (len(inst_list),))

    for idx, inst_id in enumerate(inst_list):
        inst_bgd_map = np.array(inst_map != inst_id, np.uint8)
        inst_bgd_dst = distance_transform_edt(inst_bgd_map)
        stacked_inst_bgd_dst[..., idx] = inst_bgd_dst

    near1_dst = np.amin(stacked_inst_bgd_dst, axis=2)
    near2_dst = np.expand_dims(near1_dst, axis=2)
    near2_dst = stacked_inst_bgd_dst - near2_dst
    near2_dst[near2_dst == 0] = np.PINF  # very large
    near2_dst = np.amin(near2_dst, axis=2)
    near2_dst[inst_map > 0] = 0  # the instances
    near2_dst = near2_dst + near1_dst
    # to fix pixel where near1 == near2
    near2_eve = np.expand_dims(near1_dst, axis=2)
    # to avoide the warning of a / 0
    near2_eve = (1.0 + stacked_inst_bgd_dst) / (1.0 + near2_eve)
    near2_eve[near2_eve != 1] = 0
    near2_eve = np.sum(near2_eve, axis=2)
    near2_dst[near2_eve > 1] = near1_dst[near2_eve > 1]
    #
    pix_dst = near1_dst + near2_dst
    pen_map = pix_dst / sigma
    pen_map = w0 * np.exp(-(pen_map**2) / 2)
    pen_map[inst_map > 0] = 0  # inner instances zero

    return pen_map



import torch
import torch.nn as nn

from ..utils import filter2D, gaussian_kernel2d


class WeightedBaseLoss(nn.Module):
    def __init__(
        self,
        apply_sd: bool = False,
        apply_ls: bool = False,
        apply_svls: bool = False,
        class_weights: torch.Tensor = None,
        edge_weight: float = None,
        **kwargs
    ) -> None:
        """Init a base class for weighted cross entropy based losses.

        Enables weighting for object instance edges and classes.

        Parameters
        ----------
            apply_sd : bool, default=False
                If True, Spectral decoupling regularization will be applied  to the
                loss matrix.
            apply_ls : bool, default=False
                If True, Label smoothing will be applied to the target.
            apply_svls : bool, default=False
                If True, spatially varying label smoothing will be applied to the target
            class_weights : torch.Tensor, default=None
                Class weights. A tensor of shape (C, )
            edge_weight : float, default=None
                Weight for the object instance border pixels
        """
        super().__init__()
        self.apply_sd = apply_sd
        self.apply_ls = apply_ls
        self.apply_svls = apply_svls
        self.class_weights = class_weights
        self.edge_weight = edge_weight

    def apply_spectral_decouple(
        self, loss_matrix: torch.Tensor, yhat: torch.Tensor, lam: float = 0.01
    ) -> torch.Tensor:
        """Apply spectral decoupling L2 norm after the loss.

        https://arxiv.org/abs/2011.09468

        Parameters
        ----------
            loss_matrix : torch.Tensor
                Pixelwise losses. A tensor of shape (B, H, W).
            yhat : torch.Tensor
                The pixel predictions of the model. Shape (B, C, H, W).
            lam : float, default=0.01
                Lambda constant.

        Returns
        -------
            torch.Tensor:
                SD-regularized loss matrix. Same shape as input.
        """
        # return loss_matrix + (lam / 2) * (yhat**2).mean() # which??
        return loss_matrix + (lam / 2) * (yhat**2).mean(axis=1)

    def apply_ls_to_target(
        self,
        target: torch.Tensor,
        num_classes: int,
        label_smoothing: float = 0.1,
    ) -> torch.Tensor:
        """Apply regular label smoothing to the target map.

        https://arxiv.org/abs/1512.00567

        Parameters
        ----------
            target : torch.Tensor
                The target one hot tensor. Shape (B, C, H, W). Dtype: Int64.
            num_classes : int
                Number of classes in the data.
            label_smoothing : float, default=0.1
                The smoothing coeff alpha.

        Retrurns
        --------
            Torch.Tensor:
                Label smoothed target. Same shape as input.
        """
        return target * (1 - label_smoothing) + label_smoothing / num_classes

    def apply_svls_to_target(
        self,
        target: torch.Tensor,
        num_classes: int,
        kernel_size: int = 5,
        sigma: int = 3,
        **kwargs
    ) -> torch.Tensor:
        """Apply spatially varying label smoothihng to target map.

        https://arxiv.org/abs/2104.05788

        Parameters
        ----------
            target : torch.Tensor
                The target one hot tensor. Shape (B, C, H, W). Dtype: Int64.
            num_classes : int
                Number of classes in the data.
            kernel_size : int, default=3
                Size of a square kernel.
            sigma : int, default=3
                The std of the gaussian.

        Retrurns
        --------
            Torch.Tensor:
                Label smoothed target. Same shape as input.
        """
        my, mx = kernel_size // 2, kernel_size // 2
        gaussian_kernel = gaussian_kernel2d(
            kernel_size, sigma, num_classes, device=target.device
        )
        neighborsum = (1 - gaussian_kernel[..., my, mx]) + 1e-16
        gaussian_kernel = gaussian_kernel.clone()
        gaussian_kernel[..., my, mx] = neighborsum
        svls_kernel = gaussian_kernel / neighborsum[0]

        return filter2D(target, svls_kernel) / svls_kernel[0].sum()

    def apply_class_weights(
        self, loss_matrix: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        """Multiply pixelwise loss matrix by the class weights.

        NOTE: No normalization

        Parameters
        ----------
            loss_matrix : torch.Tensor
                Pixelwise losses. A tensor of shape (B, H, W).
            target : torch.Tensor
                The target mask. Shape (B, H, W).

        Returns
        -------
            torch.Tensor:
                The loss matrix scaled with the weight matrix. Shape (B, H, W).
        """
        weight_mat = self.class_weights[target].to(target.device)  # to (B, H, W)
        loss = loss_matrix * weight_mat

        return loss

    def apply_edge_weights(
        self, loss_matrix: torch.Tensor, weight_map: torch.Tensor
    ) -> torch.Tensor:
        """Apply weights to the object boundaries.

        Basically just computes `edge_weight`**`weight_map`.

        Parameters
        ----------
            loss_matrix : torch.Tensor
                Pixelwise losses. A tensor of shape (B, H, W).
            weight_map : torch.Tensor
                Map that points to the pixels that will be weighted.
                Shape (B, H, W).

        Returns
        -------
            torch.Tensor:
                The loss matrix scaled with the nuclear boundary weights.
                Shape (B, H, W).
        """
        return loss_matrix * self.edge_weight**weight_map

    def extra_repr(self) -> str:
        """Add info to print."""
        s = "class_weights={class_weights}, edge_weight={edge_weight}"
        return s.format(**self.__dict__)



import numpy as np
import pytest
import torch

from cellseg_models_pytorch.utils import (
    ndarray_to_tensor,
    tensor_one_hot,
    tensor_to_ndarray,
    to_device,
)


@pytest.mark.parametrize(
    "in_dim_format",
    ("HW", "HWC", "BHWC", "BCHW", "BHW", pytest.param("B", marks=pytest.mark.xfail)),
)
@pytest.mark.parametrize(
    "out_dim_format",
    ("HW", "HWC", "BHWC", "BCHW", "BHW", pytest.param("B", marks=pytest.mark.xfail)),
)
def test_ndarray_to_tensor(in_dim_format, out_dim_format):
    arr = np.random.rand(14, 14)

    if in_dim_format in ("HWC", "BHWC", "BCHW"):
        arr = arr[..., None]

    if in_dim_format in ("BHWC", "BCHW", "BHW"):
        arr = arr[None, ...]

    if len(arr.shape) == 4 and in_dim_format == "BCHW":
        arr = arr.transpose(0, 3, 1, 2)

    t = ndarray_to_tensor(arr, in_dim_format, out_dim_format)
    assert isinstance(t, torch.Tensor)


@pytest.mark.parametrize(
    "in_dim_format",
    ("BCHW", "BHW", pytest.param("BHWC", marks=pytest.mark.xfail)),
)
@pytest.mark.parametrize(
    "out_dim_format",
    ("BHWC", "BHW", "HWC", "HW", pytest.param("B", marks=pytest.mark.xfail)),
)
def test_tensor_to_ndarray(in_dim_format, out_dim_format):

    if in_dim_format == "BCHW":
        arr = np.random.rand(2, 2, 14, 14)
    else:
        arr = np.random.rand(2, 14, 14)

    t = ndarray_to_tensor(arr, in_dim_format, in_dim_format)
    n = tensor_to_ndarray(t, in_dim_format, out_dim_format)
    assert isinstance(n, np.ndarray)


@pytest.mark.cuda
def test_to_device():
    tensor_cpu = torch.randint(1, 2, [3, 3])
    t = to_device(tensor_cpu)

    assert t.is_cuda


def test_tensor_one_hot(tensor_sem_map):
    t = tensor_one_hot(tensor_sem_map, n_classes=4)
    assert t.shape == torch.Size([2, 4, 6, 6])
    assert t.type() == "torch.FloatTensor"



import pytest
import torch

from cellseg_models_pytorch.losses import SSIM, CELoss, IoULoss, JointLoss, TverskyLoss
from cellseg_models_pytorch.losses.tests.test_losses import _get_dummy_pair


@torch.no_grad()
@pytest.mark.parametrize("n_classes", [1, 3])
@pytest.mark.parametrize(
    "losses",
    [
        [TverskyLoss, CELoss],
        [IoULoss],
        [SSIM, TverskyLoss, CELoss],
    ],
)
def test_joint_loss(n_classes, losses) -> None:
    yhat, target = _get_dummy_pair(n_classes)
    loss = JointLoss([l() for l in losses])
    res = loss(yhat=yhat, target=target)

    assert res.dtype == torch.float32
    assert res.size() == torch.Size([])



"""
Adapted from the diffusers repo. (Added type hints).

https://github.com/huggingface/diffusers/blob/v0.7.0/src/diffusers/models/attention.py

Copyright [2022] [huggingface]

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = ["GEGLU", "ApproximateGELU"]


class GEGLU(nn.Module):
    def __init__(self, dim_in: int, dim_out: int, **kwargs) -> None:
        """Apply a variant of the gated linear unit activation function.

        https://arxiv.org/abs/2002.05202.

        Parameters
        ----------
            dim_in : int
                The number of channels in the input.
            dim_out : int
                The number of channels in the output.
        """
        super().__init__()
        self.proj = nn.Linear(dim_in, dim_out * 2)

    def gelu(self, gate: torch.Tensor) -> torch.Tensor:
        """Run gelu activation."""
        return F.gelu(gate.to(dtype=torch.float32)).to(dtype=gate.dtype)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of he gated GELU."""
        x, gate = self.proj(x).chunk(2, dim=-1)
        return x * self.gelu(gate)


class ApproximateGELU(nn.Module):
    def __init__(self, dim_in: int, dim_out: int, **kwargs) -> None:
        """Apply the approximate form of Gaussian Error Linear Unit (GELU).

        https://arxiv.org/abs/1606.08415

        Parameters
        ----------
            dim_in : int
                The number of channels in the input.
            dim_out : int
                The number of channels in the output.
        """
        super().__init__()
        self.proj = nn.Linear(dim_in, dim_out)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the approximate GELU."""
        x = self.proj(x)
        return x * torch.sigmoid(1.702 * x)



from .attention_modules import Attention
from .base_modules import Activation, Conv, Identity, Norm, Up
from .conv_block import ConvBlock
from .conv_layer import ConvLayer
from .metaformer import MetaFormer
from .misc_modules import ChannelPool
from .self_attention_modules import SelfAttention, SelfAttentionBlock
from .token_mixers import TokenMixer, TokenMixerBlock
from .transformers import Transformer2D, TransformerLayer

__all__ = [
    "Identity",
    "Conv",
    "Up",
    "Activation",
    "Norm",
    "Attention",
    "ConvBlock",
    "ConvLayer",
    "SelfAttention",
    "SelfAttentionBlock",
    "TransformerLayer",
    "Transformer2D",
    "ChannelPool",
    "MetaFormer",
    "TokenMixer",
    "TokenMixerBlock",
]



import pytest
import torch

from cellseg_models_pytorch.modules.base_modules import Activation, Conv, Norm, Up


@pytest.mark.parametrize("activation", ["leaky-relu", "relu", "mish", "swish", None])
def test_act_forward(activation):
    act = Activation(activation)
    input = torch.rand([1, 3, 16, 16])
    output = act(input)

    assert output.dtype == input.dtype


@pytest.mark.parametrize("normalization", ["bn", "bcn", "gn", "ln2d", None])
def test_norm(normalization):
    norm = Norm(normalization, num_features=3)
    input = torch.rand([1, 3, 16, 16])
    output = norm(input)

    assert output.dtype == input.dtype


@pytest.mark.parametrize("scale_factor", [2, 4])
@pytest.mark.parametrize("upsampling", ["fixed-unpool", "bilinear", "bicubic"])
def test_up(upsampling, scale_factor):
    up = Up(upsampling, scale_factor)
    input = torch.rand([1, 3, 16, 16])
    output = up(input)

    assert output.dtype == input.dtype
    assert output.shape == torch.Size([1, 3, 16 * scale_factor, 16 * scale_factor])


@pytest.mark.parametrize("conv", ["conv", "wsconv", "scaled_wsconv"])
def test_conv(conv):
    conv = Conv(conv, in_channels=3, out_channels=3, kernel_size=3, padding=1)
    input = torch.rand([1, 3, 16, 16])
    output = conv(input)

    assert output.dtype == input.dtype
    assert output.shape == input.shape



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from pathos.multiprocessing import ThreadPool as Pool

from ...transforms.albu_transforms import IMG_TRANSFORMS, compose
from ...utils import FileHandler, TilerStitcher, fix_duplicates, remap_label

__all__ = ["BaseWriter"]

IMG_SUFFIXES = (".jpeg", ".jpg", ".tif", ".tiff", ".png")
MASK_SUFFIXES = (".mat",)


class BaseWriter(ABC):
    def __init__(
        self,
        in_dir_im: str,
        in_dir_mask: str = None,
        patch_size: Tuple[int, int] = None,
        stride: int = None,
        transforms: Optional[List[str]] = None,
        **kwargs,
    ) -> None:
        """Init base class for sliding window data writers."""
        self.stride = stride

        if isinstance(patch_size, int):
            patch_size = (patch_size, patch_size)

        self.patch_size = patch_size
        self.im_dir = Path(in_dir_im)

        # Imgs
        if not self.im_dir.exists():
            raise ValueError(f"folder: {self.im_dir} does not exist")

        if not self.im_dir.is_dir():
            raise ValueError(f"path: {self.im_dir} is not a folder")

        im_files = []
        for types in IMG_SUFFIXES:
            im_files.extend(self.im_dir.glob(f"*{types}"))
        self.fnames_im = sorted(im_files)

        # Masks
        self.mask_dir = in_dir_mask
        self.fnames_mask = None
        if in_dir_mask is not None:
            self.mask_dir = Path(in_dir_mask)

            if not self.mask_dir.exists():
                raise ValueError(f"folder: {self.mask_dir} does not exist")

            if not self.mask_dir.is_dir():
                raise ValueError(f"path: {self.mask_dir} is not a folder")

            mask_files = []
            for types in MASK_SUFFIXES:
                mask_files.extend(self.mask_dir.glob(f"*{types}"))
            self.fnames_mask = sorted(mask_files)

            if len(self.fnames_im) != len(self.fnames_mask):
                raise ValueError(
                    f"Found different number of files in {self.im_dir.as_posix()} and "
                    f"{self.mask_dir.as_posix()}."
                )

        # Transformations
        self.transforms = None
        if transforms is not None:
            allowed = list(IMG_TRANSFORMS.keys())
            if not all([tr in allowed for tr in transforms]):
                raise ValueError(
                    f"Wrong transformation. Got: {transforms}. Allowed: {allowed}."
                )

            self.transforms = compose(
                [IMG_TRANSFORMS[tr](**kwargs) for tr in transforms]
            )

    @abstractmethod
    def write(self):
        """Patch images and mask to and write them to disk."""
        raise NotImplementedError

    def get_array(
        self,
        img_path: Union[str, Path],
        mask_path: Optional[Union[str, Path]] = None,
        tiling: Optional[bool] = False,
        pre_proc: Optional[Callable] = None,
    ) -> Tuple[np.ndarray, Union[None, Dict[str, np.ndarray]]]:
        """Pipeline that (optionally) patches and transforms input images and masks.

        Parameters
        ----------
            img_path : str or Path
                Path to an image file.
            mask_path : str or Path, optional
                Path to a .mat mask file.
            tiling : bool, default=False, optional
                Flag, whether to do tiling on the images (and masks).
            pre_proc : Callable, optional
                A pre-processing function that can be used to pre-process given input
                masks before the pipeline.

        Raises
        ------
            ValueError if self.stride or self.patch_size are not set to integer values.

        Returns
        -------
            Tuple[np.ndarray, Union[None, Dict[str, np.ndarray]]]:
                The processed image & masks. If `mask_path=None`, returns no masks.
                Img shape w/o tiling: (H, W, C). Dtype: uint8.
                Img shape w/ tiling: (N, pH, pW, C). Dtype: uint8.
                Masks w/o tiling: Shapes: (H, W). Dtypes: int32.
                Masks w/ tiling: Shapes: (N, pH, pW). Dtypes: int32.
        """
        im, masks = self._read_files(img_path, mask_path, pre_proc)

        # do tiling first if flag set to True
        if tiling:
            if not isinstance(self.stride, int) and not isinstance(
                self.patch_size, int
            ):
                raise ValueError(
                    "`self.stride` and `self.patch_size` need to be integers. Got: "
                    f"self.stride={self.stride}, self.patch_size={self.patch_size}"
                )

            im, masks = self._get_tiles(im, masks)

            if masks is not None:
                if "inst_map" in masks.keys():
                    masks["inst_map"] = self._fix_instances_tiles(masks["inst_map"])

            if self.transforms is not None:
                im, masks = self._transform_tiles(im, masks)
        else:
            if masks is not None:
                if "inst_map" in masks.keys():
                    masks["inst_map"] = self._fix_instances_one(masks["inst_map"])

            if self.transforms is not None:
                im, masks = self._transform_one(im, masks)

        return im, masks

    def _fix_instances_one(self, inst_map: np.ndarray) -> np.ndarray:
        """Fix duplicate instances and remap instance labels."""
        return remap_label(fix_duplicates(inst_map))

    def _read_files(
        self,
        img_path: Union[str, Path],
        mask_path: Union[str, Path] = None,
        pre_proc: Callable = None,
    ) -> Tuple[np.ndarray, Union[None, Dict[str, np.ndarray]]]:
        """Read image and corresponding masks if there are such."""
        im = FileHandler.read_img(img_path)

        masks = None
        if mask_path is not None:
            masks = FileHandler.read_mat(mask_path, return_all=True)

            if pre_proc is not None:
                masks = pre_proc(masks)

            masks = {
                key: arr
                for key, arr in masks.items()
                if key in ("inst_map", "type_map", "sem_map")
            }

        return im, masks

    def _get_tiles(
        self,
        im: np.ndarray,
        masks: Union[Dict[str, np.ndarray], None] = None,
    ) -> Tuple[Dict[str, np.ndarray], Union[Dict[str, np.ndarray], None]]:
        """Do tiling on an image and corresponding masks if there are such."""
        # Init Tilers
        im_tiler = TilerStitcher(
            im_shape=im.shape, patch_shape=self.patch_size + (3,), stride=self.stride
        )
        im_tiles = im_tiler.patch(im)

        # Tile masks if there are such.
        mask_tiles = None
        if masks is not None:
            mask_tiles = {}
            inst = None
            types = None
            sem = None
            if "inst_map" in masks.keys():
                inst = masks["inst_map"]
            if "type_map" in masks.keys():
                types = masks["type_map"]
            if "sem_map" in masks.keys():
                sem = masks["sem_map"]

            mask_tiler = TilerStitcher(
                im_shape=inst.shape,
                patch_shape=self.patch_size + (1,),
                stride=self.stride,
            )

            if inst is not None:
                mask_tiles["inst_map"] = mask_tiler.patch(inst).squeeze()
            if types is not None:
                mask_tiles["type_map"] = mask_tiler.patch(types).squeeze()
            if sem is not None:
                mask_tiles["sem_map"] = mask_tiler.patch(sem).squeeze()

        return im_tiles, mask_tiles

    def _transform_one(
        self, im: np.ndarray, masks: Dict[str, np.ndarray] = None
    ) -> Tuple[np.ndarray, Union[Dict[str, np.ndarray], None]]:
        """Transform an image and corresponding mask if there is one."""
        if masks is not None:
            mask_names = [name for name in masks.keys()]
            masks = [mask for mask in masks.values()]
            out = self.transforms(image=im, masks=masks)
            masks = {n: mask for n, mask in zip(mask_names, out["masks"])}
        else:
            out = self.transforms(image=im)

        im = out["image"]

        return im, masks

    def _transform_tiles(
        self,
        im_tiles: np.ndarray,
        mask_tiles: Union[Dict[str, np.ndarray], None] = None,
    ) -> Tuple[np.ndarray, Union[Dict[str, np.ndarray], None]]:
        """Apply transformations to the tiles one by one."""
        n_tiles = im_tiles.shape[0]
        out_im_tiles = []

        out_mask_tiles = None
        if mask_tiles is not None:
            mask_names = [key for key in mask_tiles.keys()]
            out_mask_tiles = {k: [] for k in mask_tiles.keys()}

        for i in range(n_tiles):
            # Get one img tile
            im = im_tiles[i]

            # Get one set of mask tiles
            masks = None
            if mask_tiles is not None:
                masks = {n: mask_tiles[n][i] for n in mask_names}

            # transform imgs & masks
            im_tr, masks_tr = self._transform_one(im, masks)

            out_im_tiles.append(im_tr)
            if mask_tiles is not None:
                for mask_name in mask_names:
                    out_mask_tiles[mask_name].append(masks_tr[mask_name])

        # convert list of 2D-arrays to np.ndarray
        out_im_tiles = np.array(out_im_tiles)
        if mask_tiles is not None:
            for k, arr in out_mask_tiles.items():
                out_mask_tiles[k] = np.array(arr)

        return out_im_tiles, out_mask_tiles

    def _fix_instances_tiles(self, patches_inst: np.ndarray) -> np.ndarray:
        """Fix repeated labels and remap them in a patched instance labelled mask."""
        insts = []

        for i in range(patches_inst.shape[0]):
            insts.append(self._fix_instances_one(patches_inst[i]))

        insts = np.array(insts)

        return insts.astype("int32")

    def _write_parallel(self, write_func: Callable, arg_list: List) -> None:
        """Write patches in parallel."""
        with Pool() as pool:
            it = pool.imap(write_func, arg_list)
            for _ in it:
                pass

    def _write_sequential(self, write_func: Callable, arg_list: List) -> None:
        """Write patches sequentially."""
        for kwargs in arg_list:
            write_func(**kwargs)



from .cellpose import gen_flow_maps
from .contour import gen_contour_maps
from .dist import gen_dist_maps
from .edge_weights import gen_weight_maps
from .hovernet import gen_hv_maps
from .omnipose import gen_omni_flow_maps, smooth_distance
from .stardist import evenly_spaced_points, gen_radial_distmaps, gen_stardist_maps

__all__ = [
    "evenly_spaced_points",
    "gen_radial_distmaps",
    "gen_stardist_maps",
    "gen_flow_maps",
    "gen_hv_maps",
    "gen_omni_flow_maps",
    "smooth_distance",
    "gen_dist_maps",
    "gen_contour_maps",
    "gen_weight_maps",
]



from .ce import CELoss
from .dice import DiceLoss
from .focal import FocalLoss
from .grad_mse import GradMSE
from .iou import IoULoss
from .mse import MSE
from .sce import SCELoss
from .ssim import MSSSIM, SSIM
from .tversky import TverskyLoss

SEG_LOSS_LOOKUP = {
    "iou": IoULoss,
    "dice": DiceLoss,
    "tversky": TverskyLoss,
    "ce": CELoss,
    "sce": SCELoss,
    "focal": FocalLoss,
    "mse": MSE,
    "gmse": GradMSE,
    "ssim": SSIM,
    "msssim": MSSSIM,
}


__all__ = [
    "SEG_LOSS_LOOKUP",
    "REG_LOSS_LOOKUP",
    "MSE",
    "GradMSE",
    "SSIM",
    "MSSSIM",
    "DiceLoss",
    "IoULoss",
    "TverskyLoss",
    "FocalLoss",
    "CELoss",
    "SCELoss",
]



"""
Ported from the hovernet repo with minor mods.

MIT License

Copyright (c) 2020 vqdang

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import numpy as np
from scipy.ndimage import distance_transform_cdt, distance_transform_edt

from cellseg_models_pytorch.utils import bounding_box

__all__ = ["gen_dist_maps"]


def gen_dist_maps(
    inst_map: np.ndarray, normalize: bool = True, euclidean: bool = True
) -> np.ndarray:
    """Compute distance transforms for every labelled object.

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).
        normalize : bool, default=True
            Normalize the distance maps to [0, 1].
        euclidean : bool, default=True
            If True, uses the euclidean transformation. Else, the chessboard one.

    Returns
    -------
        np.ndarray:
            Distance map of the labelled objects. Shape (H, W).
    """
    dist = np.zeros_like(inst_map, dtype=np.float64)

    inst_list = list(np.unique(inst_map))
    if 0 in inst_list:
        inst_list.remove(0)

    for inst_id in inst_list:
        inst = np.array(inst_map == inst_id, np.uint8)

        y1, y2, x1, x2 = bounding_box(inst)
        y1 = y1 - 2 if y1 - 2 >= 0 else y1
        x1 = x1 - 2 if x1 - 2 >= 0 else x1
        x2 = x2 + 2 if x2 + 2 <= inst_map.shape[1] - 1 else x2
        y2 = y2 + 2 if y2 + 2 <= inst_map.shape[0] - 1 else y2

        inst = inst[y1:y2, x1:x2]

        if inst.shape[0] < 2 or inst.shape[1] < 2:
            continue

        # chessboard distance map generation
        # normalize distance to 0-1
        if euclidean:
            inst_dist = distance_transform_edt(inst)
        else:
            inst_dist = distance_transform_cdt(inst)
        inst_dist = inst_dist.astype("float64")

        if normalize:
            max_value = np.amax(inst_dist)
            if max_value <= 0:
                continue
            inst_dist = inst_dist / (np.max(inst_dist) + 1e-10)

        dist_map_box = dist[y1:y2, x1:x2]
        dist_map_box[inst > 0] = inst_dist[inst > 0]

    return dist



import numpy as np
import skimage.filters as filters
import skimage.segmentation as segm

from .mask_utils import remove_debris_binary

__all__ = [
    "naive_thresh_prob",
    "naive_thresh",
    "niblack_thresh",
    "sauvola_thresh",
    "morph_chan_vese_thresh",
    "argmax",
]


def naive_thresh_prob(
    prob_map: np.ndarray, threshold: float = 0.5, **kwargs
) -> np.ndarray:
    """Threshold a sigmoid/softmax activated soft mask.

    Parameters
    ----------
        prob_map : np.ndarray
            Soft mask to be thresholded. Shape (H, W)
        threshold: float, default=0.5
            Thresholding cutoff between [0, 1]

    Returns
    -------
        np.ndarray:
            Thresholded soft mask. Shape: (H, W). Type: uint8.
    """
    if not 0 <= threshold <= 1:
        raise ValueError(f"thresh = {threshold}. given threshold not between [0,1]")

    seg = prob_map.copy()
    seg = seg >= threshold
    seg = remove_debris_binary(seg)

    return seg


def naive_thresh(prob_map: np.ndarray, threshold: int = 2, **kwargs) -> np.ndarray:
    """Threshold a soft mask. Values can be logits or probabilites.

    Parameters
    ----------
        prob_map : np.ndarray
            Soft mask to be thresholded. Shape (H, W).
        threshold : int, default=2
            Value used to divide the max value of the mask.

    Returns
    -------
        np.ndarray:
            Thresholded soft mask. Shape: (H, W). Type: uint8.
    """
    seg = prob_map.copy()
    seg[seg < np.amax(prob_map) / threshold] = 0
    seg[seg > np.amax(prob_map) / threshold] = 1
    seg = remove_debris_binary(seg)

    return seg


def niblack_thresh(prob_map: np.ndarray, win_size: int = 13, **kwargs) -> np.ndarray:
    """Do niblack thresholding (skimage wrapper).

    Parameters
    ----------
        prob_map : np.ndarray
            Soft mask to be thresholded. Shape (H, W).
        win_size : int, default=13
            Size of the window used in the thresholding algorithm.

    Returns
    -------
        np.ndarray:
            Thresholded soft mask. Shape: (H, W). Type: uint8.
    """
    thresh = filters.threshold_niblack(prob_map, window_size=win_size)
    seg = prob_map > thresh
    seg = remove_debris_binary(seg)

    return seg


def sauvola_thresh(prob_map: np.ndarray, win_size: int = 33, **kwargs) -> np.ndarray:
    """Do sauvola thresholding (skimage wrapper).

    Parameters
    ----------
        prob_map : np.ndarray
            Soft mask to be thresholded. Shape (H, W).
        win_size : int, default=33
            Size of the window used in the thresholding algorithm.

    Returns
    -------
        np.ndarray:
            Thresholded soft mask. Shape: (H, W). Type: uint8.
    """
    thresh = filters.threshold_sauvola(prob_map, window_size=win_size)
    seg = prob_map > thresh
    seg = remove_debris_binary(seg)

    return seg


def morph_chan_vese_thresh(prob_map: np.ndarray, **kwargs) -> np.ndarray:
    """Morphological chan vese method for thresholding. Skimage wrapper.

    Parameters
    ----------
        prob_map : np.ndarray
            Soft mask to be thresholded. Shape (H, W).

    Returns
    -------
        np.ndarray:
            Thresholded soft mask. Shape: (H, W). Type: uint8.
    """
    init_ls = segm.checkerboard_level_set(prob_map.shape, 2)
    ls = segm.morphological_chan_vese(prob_map, 35, smoothing=1, init_level_set=init_ls)

    hist = np.histogram(ls)[0]
    if hist[-1] > hist[0]:
        ls = 1 - ls

    seg = remove_debris_binary(ls)

    return seg


def argmax(prob_map: np.ndarray, **kwargs) -> np.ndarray:
    """Take argmax of a one_hot logits or prob map.

    Parameters
    ----------
        prob_map : np.ndarray
            The probability map of shape (H, W, C)|(H, W).

    Returns
    -------
        np.ndarray:
            a mask of argmax indices. Shape: (H, W). Type: uint8.
    """
    if len(prob_map.shape) == 2:
        inv_prob = 1 - prob_map
        prob_map = np.stack([inv_prob, prob_map], axis=-1)

    seg = np.argmax(prob_map, axis=-1).astype("u4")

    return seg



from typing import Callable, Dict, List, Tuple, Union

import numpy as np

from .binary_metrics import get_stats, iou_score

__all__ = [
    "pairwise_pixel_stats",
    "pairwise_object_stats",
    "panoptic_quality",
    "average_precision",
    "aggregated_jaccard_index",
    "dice2",
    "iou_multiclass",
    "dice_multiclass",
    "f1score_multiclass",
    "accuracy_multiclass",
    "sensitivity_multiclass",
    "specificity_multiclass",
]


def pairwise_pixel_stats(
    true: np.ndarray,
    pred: np.ndarray,
    num_classes: int = None,
    metric_func: Callable = None,
) -> Union[List[np.ndarray], None]:
    """Compute the # of TP, FP, FN pixels for each object in a labelled/semantic mask.

    Optionally a binary metric can be computed instead of the satistics.
    Atleast 2x faster than computing with `np.histogram2d`.

    Parameters
    ----------
        true : np.ndarray
            Ground truth (semantic or labelled mask). Shape (H, W).
        pred : np.ndarray
            Predicted (semantic or labelled mask). Shape (H, W).
        num_classes : int, optional
            Number of classes in the dataset. If None, stats are computed for instances.
            If not None stats are computed for classes i.e. semantic segmentation masks.
        metric_func : Callable, optional
            A binary metric function. e.g. `iou_score` or `dice`.

    Returns
    -------
        List[np.ndarray, ...] or None:
            A List of 2D arrays (i, j) where i corresponds to a ground
            truth label and j corresponds to a predicted label. Each value
            of the matrix is the computed statistic or metric at pos (i, j).
            By default. returns the tp, fp, and fn matrices.

            If stats computed for instances:
                Shape: (n_labels_gt, n_labels_pred). Dtype. float64.
            If stats computed for classes:
                Shape: (num_classes, num_classes). Dtype. float64.
    """
    if num_classes is not None:
        true_labels = list(range(num_classes))
        pred_labels = list(range(num_classes))
    else:
        true_labels = list(np.unique(true))[1:]
        pred_labels = list(np.unique(pred))[1:]

    true_objects = {}
    for t in true_labels:
        true_obj = np.array(true == t, np.uint8)
        true_objects[t] = true_obj

    pred_objects = {}
    for p in pred_labels:
        pred_obj = np.array(pred == p, np.uint8)
        pred_objects[p] = pred_obj

    # array dims
    i = len(true_labels)
    j = len(pred_labels)

    # init return list
    ret = []
    if i > 0 and j > 0:
        ret.append(np.zeros((i, j), dtype=np.float64))
        if metric_func is None:
            ret.append(np.zeros((i, j), dtype=np.float64))
            ret.append(np.zeros((i, j), dtype=np.float64))

    for true_label in true_labels:
        true_obj = true_objects[true_label]

        overlap = pred[true_obj > 0]
        overlap_label = np.unique(overlap)
        for pred_label in overlap_label:

            # ignore bg and empty preds in instance mode
            if pred_label == 0 and num_classes is None:
                continue

            pred_obj = pred_objects[pred_label]
            tp, fp, fn = get_stats(true_obj, pred_obj)

            if num_classes is None:
                ix = true_label - 1
                jx = pred_label - 1
            else:
                ix = true_label
                jx = pred_label

            # compute a metric or add stats
            if metric_func is not None:
                ret[0][ix, jx] = metric_func(tp, fp, fn)
            else:
                ret[0][ix, jx] = tp.sum()
                ret[1][ix, jx] = fp.sum()
                ret[2][ix, jx] = fn.sum()

    return ret


def pairwise_object_stats(
    matches: np.ndarray, sum_reduce: bool = True
) -> Union[Tuple[int, int, int], Tuple[List[bool]]]:
    """Compute the TP, FP, FN objects from a boolean contigency table.

    Parameters
    ----------
        matches : np.ndarray
            A pairwise boolean matrix where True values at pos (i, j)
            indicate correctly detected objects for the corresponding
            labels i and j. Shape: (n_labels_gt, n_labels_pred).
        sum_reduce : bool, default=True
            Reduce the boolean indice arrays by summing to get the correct
            number of TP, FP, and FN objects.

    Returns
    -------
        Tuple[int, int, int]:
            The number of TP objects, FP objects, and FN objects in
            a labelled mask.

    """
    true_hits = matches.sum(axis=0)
    pred_hits = matches.sum(axis=1)

    tp_objects = pred_hits >= 1  # indices of correctly predicted objects
    fp_objects = pred_hits == 0  # indices of missed objects in prediciton
    fn_objects = true_hits == 0  # indices of extra objects in prediction

    if sum_reduce:
        tp_objects = tp_objects.sum()
        fp_objects = fp_objects.sum()
        fn_objects = fn_objects.sum()

    return tp_objects, fp_objects, fn_objects


def panoptic_quality(
    true: np.ndarray, pred: np.ndarray, thresh: float = 0.5, eps: float = 1e-8
) -> Dict[str, float]:
    """Compute the panoptic quality of a lebelled mask.

    Parameters
    ----------
        true : np.ndarray
            Ground truth (labelled mask). Shape (H, W).
        pred : np.ndarray
            Predicted (labelled mask). Shape (H, W).
        thresh : float, default=0.5
            Threshold for the iou to include the prediction as TP
        eps : float, default=1e-8:
            Epsilon to avoid zero div errors.

    Returns
    -------
        Dict[str, float]:
            Dictionary containing the detection quality (dq), segmentation
            quality (sq) and panoptic quality (pq) values.
    """
    iou = pairwise_pixel_stats(true, pred, metric_func=iou_score)
    res = {"pq": 0.0, "sq": 0.0, "dq": 0.0}

    if iou:
        iou = iou[0]
        matches = iou > thresh
        tp_objects, fp_objects, fn_objects = pairwise_object_stats(matches)

        dq = tp_objects / (tp_objects + 0.5 * fp_objects + 0.5 * fn_objects + eps)
        sq = iou[matches].sum() / (tp_objects + eps)
        pq = dq * sq

        res["pq"] = pq
        res["sq"] = sq
        res["dq"] = dq

    return res


def average_precision(
    true: np.ndarray, pred: np.ndarray, thresh: float = 0.5, eps: float = 1e-8
) -> float:
    """Compute the average precision of a labelled mask.

    Parameters
    ----------
        true : np.ndarray
            Ground truth (labelled mask). Shape (H, W).
        pred : np.ndarray
            Predicted (labelled mask). Shape (H, W).
        thresh : float, default=0.5
            Threshold for the iou to include the prediction as TP
        eps : float, default=1e-8:
            Epsilon to avoid zero div errors.

    Returns
    -------
        float:
            The computed precision.
    """
    iou = pairwise_pixel_stats(pred, true, metric_func=iou_score)
    ap = 0.0

    if iou:
        iou = iou[0]
        matches = iou > thresh
        tp_objects, fp_objects, _ = pairwise_object_stats(matches)

        ap = tp_objects / (tp_objects + fp_objects + eps)

    return ap


def dice2(true: np.ndarray, pred: np.ndarray, eps: float = 1e-8) -> float:
    """Compute the DICE2 metric for a labelled mask.

    Parameters
    ----------
        true : np.ndarray
            Ground truth (labelled mask). Shape (H, W).
        pred : np.ndarray
            Predicted (labelled mask). Shape (H, W).
        eps : float, default=1e-8:
            Epsilon to avoid zero div errors.

    Returns
    -------
        float:
            The computed dice2 metric.

    """
    dice2 = 0.0
    stats = pairwise_pixel_stats(true, pred)

    if stats:
        tp, fp, fn = stats

        numerator = 2 * tp[tp > 0].sum()
        denominator = numerator + fp[fp > 0].sum() + fn[fn > 0].sum() + eps
        dice2 = numerator / denominator

    return dice2


def aggregated_jaccard_index(
    true: np.ndarray, pred: np.ndarray, thresh: float = 0.5, eps: float = 1e-8
) -> float:
    """Compute the aggregated jaccard index (AJI) for a labelled mask.

    Parameters
    ----------
        true : np.ndarray
            Ground truth (labelled mask). Shape (H, W).
        pred : np.ndarray
            Predicted (labelled mask). Shape (H, W).
        thresh : float, default=0.5
            Threshold for the iou to include the prediction as TP
        eps : float, default=1e-8:
            Epsilon to avoid zero div errors.

    Returns
    -------
        float:
            The computed aji.
    """
    aji = 0.0
    stats = pairwise_pixel_stats(true, pred)

    if stats:
        tp, fp, fn = stats

        inter = tp
        union = tp + fp + fn

        # Get the number of pixels from the matched objects
        matches = inter == np.amax(inter, axis=1, keepdims=True, initial=1e-8)
        inter = inter[matches].sum()
        union = union[matches].sum()

        # Get the num of pixels from the missed objects
        _, fp_objects, fn_objects = pairwise_object_stats(matches, sum_reduce=False)
        unpaired_true_labels = np.nonzero(fp_objects)[0] + 1
        unpaired_pred_labels = np.nonzero(fn_objects)[0] + 1

        for true_id in unpaired_true_labels:
            union += (true == true_id).sum()
        for pred_id in unpaired_pred_labels:
            union += (pred == pred_id).sum()

        # compute aji
        aji = inter / (union + eps)

    return aji


def _absent_inds(true: np.ndarray, pred: np.ndarray, num_classes: int) -> np.ndarray:
    """Get the class indices that are not present in either `true` or `pred`."""
    t = np.unique(true)
    p = np.unique(pred)
    not_pres = np.setdiff1d(np.arange(num_classes), np.union1d(t, p))

    return not_pres


def iou_multiclass(
    true: np.ndarray,
    pred: np.ndarray,
    num_classes: int,
    eps: float = 1e-8,
    clamp_absent: bool = True,
) -> np.ndarray:
    """Compute multi-class intersection over union for semantic segmentation masks.

    Parameters
    ----------
        true : np.ndarray
            Ground truth semantic mask. Shape (H, W).
        pred : np.ndarray
            Predicted semantic mask. Shape (H, W).
        num_classes : int
            Number of classes in the training dataset.
        eps : float, default=1e-8:
            Epsilon to avoid zero div errors.
        clamp_absent : bool, default=True
            If a class is not present in either true or pred, the value of that ix
            in the result array will be clamped to -1.0.

    Returns
    -------
        np.ndarray:
            Per class IoU-metrics. Shape: (num_classes,).
    """
    tp, fp, fn = pairwise_pixel_stats(true, pred, num_classes=num_classes)
    tp = tp.diagonal()
    fp = fp.diagonal()
    fn = fn.diagonal()

    iou = tp / (tp + fp + fn + eps)

    if clamp_absent:
        not_pres = _absent_inds(true, pred, num_classes)
        iou[not_pres] = -1.0

    return iou


def accuracy_multiclass(
    true: np.ndarray,
    pred: np.ndarray,
    num_classes: int,
    eps: float = 1e-8,
    clamp_absent: bool = True,
) -> np.ndarray:
    """Compute multi-class accuracy for semantic segmentation masks.

    Parameters
    ----------
        true : np.ndarray
            Ground truth semantic mask. Shape (H, W).
        pred : np.ndarray
            Predicted semantic mask. Shape (H, W).
        num_classes : int
            Number of classes in the training dataset.
        eps : float, default=1e-8:
            Epsilon to avoid zero div errors.
        clamp_absent: bool = True
            If a class is not present in either true or pred, the value of that ix
            in the result array will be clamped to -1.0.

    Returns
    -------
        np.ndarray:
            Per class accuracy-metrics. Shape: (num_classes,).
    """
    tp, fp, fn = pairwise_pixel_stats(true, pred, num_classes=num_classes)
    tp = tp.diagonal()
    fp = fp.diagonal()
    fn = fn.diagonal()
    tn = np.prod(true.shape) - (tp + fn + fp)

    accuracy = (tp + tn) / (tp + fp + fn + tn + eps)

    if clamp_absent:
        not_pres = _absent_inds(true, pred, num_classes)
        accuracy[not_pres] = -1.0

    return accuracy


def f1score_multiclass(
    true: np.ndarray,
    pred: np.ndarray,
    num_classes: int,
    eps: float = 1e-8,
    clamp_absent: bool = True,
) -> np.ndarray:
    """Compute multi-class f1-score for semantic segmentation masks.

    Parameters
    ----------
        true : np.ndarray
            Ground truth semantic mask. Shape (H, W).
        pred : np.ndarray
            Predicted semantic mask. Shape (H, W).
        num_classes : int
            Number of classes in the training dataset.
        eps : float, default=1e-8:
            Epsilon to avoid zero div errors.
        clamp_absent: bool = True
            If a class is not present in either true or pred, the value of that ix
            in the result array will be clamped to -1.0.

    Returns
    -------
        np.ndarray:
            Per class f1score-metrics. Shape: (num_classes,).
    """
    tp, fp, fn = pairwise_pixel_stats(true, pred, num_classes=num_classes)
    tp = tp.diagonal()
    fp = fp.diagonal()
    fn = fn.diagonal()

    f1 = tp / (0.5 * fp + 0.5 * fn + tp + eps)

    if clamp_absent:
        not_pres = _absent_inds(true, pred, num_classes)
        f1[not_pres] = -1.0

    return f1


def dice_multiclass(
    true: np.ndarray,
    pred: np.ndarray,
    num_classes: int,
    eps: float = 1e-8,
    clamp_absent: bool = True,
) -> np.ndarray:
    """Compute multi-class dice for semantic segmentation masks.

    Parameters
    ----------
        true : np.ndarray
            Ground truth semantic mask. Shape (H, W).
        pred : np.ndarray
            Predicted semantic mask. Shape (H, W).
        num_classes : int
            Number of classes in the training dataset.
        eps : float, default=1e-8:
            Epsilon to avoid zero div errors.
        clamp_absent: bool = True
            If a class is not present in either true or pred, the value of that ix
            in the result array will be clamped to -1.0.

    Returns
    -------
        np.ndarray:
            Per class dice-metrics. Shape: (num_classes,).
    """
    tp, fp, fn = pairwise_pixel_stats(true, pred, num_classes=num_classes)
    tp = tp.diagonal()
    fp = fp.diagonal()
    fn = fn.diagonal()

    dice = 2 * tp / (2 * tp + fp + fn + eps)

    if clamp_absent:
        not_pres = _absent_inds(true, pred, num_classes)
        dice[not_pres] = -1.0

    return dice


def sensitivity_multiclass(
    true: np.ndarray,
    pred: np.ndarray,
    num_classes: int,
    eps: float = 1e-8,
    clamp_absent: bool = True,
) -> np.ndarray:
    """Compute multi-class sensitivity for semantic segmentation masks.

    Parameters
    ----------
        true : np.ndarray
            Ground truth semantic mask. Shape (H, W).
        pred : np.ndarray
            Predicted semantic mask. Shape (H, W).
        num_classes : int
            Number of classes in the training dataset.
        eps : float, default=1e-8:
            Epsilon to avoid zero div errors.
        clamp_absent: bool = True
            If a class is not present in either true or pred, the value of that ix
            in the result array will be clamped to -1.0.

    Returns
    -------
        np.ndarray:
            Per class sensitivity-metrics. Shape: (num_classes,).
    """
    tp, fp, fn = pairwise_pixel_stats(true, pred, num_classes=num_classes)
    tp = tp.diagonal()
    fp = fp.diagonal()
    fn = fn.diagonal()

    sensitivity = tp / (tp + fn + eps)

    if clamp_absent:
        not_pres = _absent_inds(true, pred, num_classes)
        sensitivity[not_pres] = -1.0

    return sensitivity


def specificity_multiclass(
    true: np.ndarray,
    pred: np.ndarray,
    num_classes: int,
    eps: float = 1e-8,
    clamp_absent: bool = True,
) -> np.ndarray:
    """Compute multi-class specificity for semantic segmentation masks.

    Parameters
    ----------
        true : np.ndarray
            Ground truth semantic mask. Shape (H, W).
        pred : np.ndarray
            Predicted semantic mask. Shape (H, W).
        num_classes : int
            Number of classes in the training dataset.
        eps : float, default=1e-8:
            Epsilon to avoid zero div errors.
        clamp_absent: bool = True
            If a class is not present in either true or pred, the value of that ix
            in the result array will be clamped to -1.0.

    Returns
    -------
        np.ndarray:
            Per class specificity-metrics. Shape: (num_classes,).
    """
    tp, fp, fn = pairwise_pixel_stats(true, pred, num_classes=num_classes)
    tp = tp.diagonal()
    fp = fp.diagonal()
    fn = fn.diagonal()

    specificity = tp / (tp + fp + eps)

    if clamp_absent:
        not_pres = _absent_inds(true, pred, num_classes)
        specificity[not_pres] = -1.0

    return specificity



try:
    from ._composition import apply_each, compose, to_tensor, to_tensorv3
    from .img_transforms import (
        blur_transforms,
        center_crop,
        hue_saturation_transforms,
        non_rigid_transforms,
        non_spatial_transforms,
        random_crop,
        resize,
        rigid_transforms,
    )
    from .inst_transforms import (
        binarize_transform,
        cellpose_transform,
        contour_transform,
        dist_transform,
        edgeweight_transform,
        hovernet_transform,
        omnipose_transform,
        smooth_dist_transform,
        stardist_opt_transform,
        stardist_transform,
    )
    from .norm_transforms import (
        imgnorm_transform,
        minmaxnorm_transform,
        percentilenorm_transform,
    )

    IMG_TRANSFORMS = {
        "blur": blur_transforms,
        "center_crop": center_crop,
        "hue_sat": hue_saturation_transforms,
        "non_rigid": non_rigid_transforms,
        "rigid": rigid_transforms,
        "non_spatial": non_spatial_transforms,
        "resize": resize,
        "random_crop": random_crop,
    }

    INST_TRANSFORMS = {
        "cellpose": cellpose_transform,
        "omnipose": omnipose_transform,
        "hovernet": hovernet_transform,
        "stardist": stardist_transform,
        "stardist_opt": stardist_opt_transform,
        "contour": contour_transform,
        "dist": dist_transform,
        "edgeweight": edgeweight_transform,
        "smooth_dist": smooth_dist_transform,
        "binarize": binarize_transform,
    }

    NORM_TRANSFORMS = {
        "norm": imgnorm_transform,
        "percentile": percentilenorm_transform,
        "minmax": minmaxnorm_transform,
    }

    __all__ = [
        "IMG_TRANSFORMS",
        "NORM_TRANSFORMS",
        "INST_TRANSFORMS",
        "rigid_transforms",
        "non_rigid_transforms",
        "non_spatial_transforms",
        "hue_saturation_transforms",
        "blur_transforms",
        "random_crop",
        "center_crop",
        "resize",
        "compose",
        "apply_each",
        "to_tensor",
        "to_tensorv3",
        "cellpose_transform",
        "omnipose_transform",
        "hovernet_transform",
        "stardist_transform",
        "stardist_orig_transform",
        "dist_transform",
        "smooth_dist_transform",
        "contour_transform",
        "edgeweight_transform",
        "binarize_transform",
        "imgnorm_transform",
        "percentilenorm_transform",
        "minmaxnorm_transform",
    ]
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "To use the `csmp.transforms.albu_transforms` module, "
        "the albumentations lib is needed. Install with `pip install albumentations`"
    )



from .file_manager import FileHandler
from .img_utils import (
    float2ubyte,
    minmax_normalize,
    normalize,
    percentile_normalize,
    percentile_normalize99,
)
from .mask_utils import (
    binarize,
    bounding_box,
    center_crop,
    draw_stuff_contours,
    draw_thing_contours,
    fill_holes_semantic,
    fix_duplicates,
    get_inst_centroid,
    get_inst_types,
    get_type_instances,
    intersection,
    label_semantic,
    majority_vote_parallel,
    majority_vote_sequential,
    med_filt_parallel,
    med_filt_sequential,
    one_hot,
    remap_label,
    remove_1px_boundary,
    remove_debris_binary,
    remove_debris_instance,
    remove_debris_semantic,
    remove_small_objects,
    soft_type_flatten,
    type_map_flatten,
)
from .multiproc import run_pool
from .patching import (
    TilerStitcher,
    TilerStitcherTorch,
    _get_margins,
    extract_patches_numpy,
    extract_patches_torch,
    stitch_patches_numpy,
    stitch_patches_torch,
)
from .tensor_img_utlls import (
    NORM_LOOKUP,
    dataset_normalize_torch,
    minmax_normalize_torch,
    normalize_torch,
    percentile,
    percentile_normalize_torch,
)
from .tensor_kernels import filter2D, gaussian, gaussian_kernel2d, sobel_hv
from .tensor_utils import (
    ndarray_to_tensor,
    tensor_one_hot,
    tensor_to_ndarray,
    to_device,
)
from .thresholding import (
    argmax,
    morph_chan_vese_thresh,
    naive_thresh,
    naive_thresh_prob,
    niblack_thresh,
    sauvola_thresh,
)

THRESH_LOOKUP = {
    "argmax": argmax,
    "naive": naive_thresh_prob,
    "sauvola": sauvola_thresh,
    "niblack": niblack_thresh,
}


__all__ = [
    "THRESH_LOOKUP",
    "FileHandler",
    "percentile_normalize",
    "percentile_normalize99",
    "normalize",
    "minmax_normalize",
    "float2ubyte",
    "remove_small_objects",
    "binarize",
    "fix_duplicates",
    "remove_1px_boundary",
    "remap_label",
    "center_crop",
    "bounding_box",
    "get_inst_types",
    "get_inst_centroid",
    "get_type_instances",
    "one_hot",
    "type_map_flatten",
    "soft_type_flatten",
    "remove_debris_binary",
    "remove_debris_instance",
    "remove_debris_semantic",
    "fill_holes_semantic",
    "label_semantic",
    "naive_thresh_prob",
    "naive_thresh",
    "niblack_thresh",
    "sauvola_thresh",
    "morph_chan_vese_thresh",
    "argmax",
    "TilerStitcher",
    "TilerStitcherTorch",
    "extract_patches_numpy",
    "stitch_patches_numpy",
    "extract_patches_torch",
    "stitch_patches_torch",
    "_get_margins",
    "gaussian",
    "gaussian_kernel2d",
    "sobel_hv",
    "filter2D",
    "ndarray_to_tensor",
    "tensor_to_ndarray",
    "to_device",
    "tensor_one_hot",
    "normalize_torch",
    "minmax_normalize_torch",
    "percentile",
    "percentile_normalize_torch",
    "dataset_normalize_torch",
    "NORM_LOOKUP",
    "draw_stuff_contours",
    "draw_thing_contours",
    "majority_vote_sequential",
    "majority_vote_parallel",
    "med_filt_parallel",
    "med_filt_sequential",
    "intersection",
    "run_pool",
]



import pytest

from cellseg_models_pytorch.inference import ResizeInferer, SlidingWindowInferer
from cellseg_models_pytorch.models import cellpose_plus


@pytest.mark.parametrize("batch_size", [1, 2])
def test_slidingwin_inference(img_dir, batch_size):
    model = cellpose_plus(sem_classes=3, type_classes=3, long_skip="unet")

    inferer = SlidingWindowInferer(
        model,
        img_dir,
        out_activations={"sem": "softmax", "type": "softmax", "cellpose": "tanh"},
        out_boundary_weights={"sem": False, "type": False, "cellpose": True},
        patch_size=(256, 256),
        stride=256,
        padding=80,
        instance_postproc="hovernet",
        batch_size=batch_size,
        save_intermediate=False,
        device="cpu",
        parallel=False,
    )

    inferer.infer()

    samples = list(inferer.out_masks.keys())
    assert inferer.out_masks[samples[0]]["inst"].shape == (512, 512)


@pytest.mark.parametrize("batch_size", [1, 2])
def test_resize_inference(img_dir, batch_size):
    model = cellpose_plus(sem_classes=3, type_classes=3, long_skip="unet")

    inferer = ResizeInferer(
        model,
        img_dir,
        out_activations={"sem": "softmax", "type": "softmax", "cellpose": "tanh"},
        out_boundary_weights={"sem": False, "type": False, "cellpose": True},
        resize=(256, 256),
        padding=80,
        instance_postproc="hovernet",
        batch_size=batch_size,
        save_intermediate=False,
        device="cpu",
        parallel=False,
    )

    inferer.infer()

    samples = list(inferer.out_masks.keys())
    assert inferer.out_masks[samples[0]]["inst"].shape == (512, 512)



from itertools import permutations

from .criterions import (
    MSE,
    MSSSIM,
    SEG_LOSS_LOOKUP,
    SSIM,
    CELoss,
    DiceLoss,
    FocalLoss,
    GradMSE,
    IoULoss,
    TverskyLoss,
)
from .joint_loss import JointLoss
from .loss_module import Loss
from .multitask_loss import MultiTaskLoss

JOINT_SEG_LOSSES = []
for i in range(1, 5):
    JOINT_SEG_LOSSES.extend(
        ["_".join(t) for t in permutations(SEG_LOSS_LOOKUP.keys(), i)]
    )

__all__ = [
    "SEG_LOSS_LOOKUP",
    "JOINT_SEG_LOSSES",
    "Loss",
    "JointLoss",
    "MultiTaskLoss",
    "MSE",
    "MSSSIM",
    "CELoss",
    "SSIM",
    "GradMSE",
    "TverskyLoss",
    "FocalLoss",
    "DiceLoss",
    "IoULoss",
]



from .attention_ops import compute_mha, mha, slice_mha
from .exact_attention import ExactSelfAttention
from .linformer import LinformerAttention

SELFATT_LOOKUP = {
    "exact": ExactSelfAttention,
    "linformer": LinformerAttention,
}

__all__ = [
    "ExactSelfAttention",
    "LinformerAttention",
    "mha",
    "slice_mha",
    "compute_mha",
    "SELFATT_LOOKUP",
]



import itertools
from pathlib import Path
from typing import Any, Dict, Iterable, List, Tuple, Union

import numpy as np
from tqdm import tqdm

from ..inference import BaseInferer
from ..metrics import (
    accuracy_multiclass,
    aggregated_jaccard_index,
    average_precision,
    dice2,
    dice_multiclass,
    f1score_multiclass,
    iou_multiclass,
    panoptic_quality,
    sensitivity_multiclass,
    specificity_multiclass,
)
from .file_manager import FileHandler
from .mask_utils import get_type_instances, remap_label
from .multiproc import run_pool

__all__ = ["SegBenchmarker"]


INST_METRIC_LOOKUP = {
    "pq": panoptic_quality,
    "aji": aggregated_jaccard_index,
    "dice2": dice2,
    "ap": average_precision,
}


SEM_METRIC_LOOKUP = {
    "iou": iou_multiclass,
    "dice": dice_multiclass,
    "f1": f1score_multiclass,
    "acc": accuracy_multiclass,
    "sensitivity": sensitivity_multiclass,
    "specificity": specificity_multiclass,
}


class SegBenchmarker:
    def __init__(
        self,
        true_path: str,
        pred_path: str = None,
        inferer: BaseInferer = None,
        type_classes: Dict[str, int] = None,
        sem_classes: Dict[str, int] = None,
    ) -> None:
        """Run benchmarking, given prediction and ground truth mask folders or hdf5 dbs.

        NOTE: Can also take in an Inferer object that contains predictions.

        Parameters
        ----------
            true_path : str
                Path to the ground truth .mat files or h5 db. The gt files have to have
                matching names to the pred filenames.
            pred_path : str, optional
                Path to thse prediction .mat files. The pred files have to have matching
                names to the gt filenames. If None, the inferer object storing the
                predictions will be used instead.
            inferer : BaseInferer, optional
                Infere object storing predictions of a model. If None, the `pred_path`
                will be used to load the predictions instead.
            type_classes : Dict[str, int], optional
                Cell type class dict. E.g. {"bg": 0, "epithelial": 1, "immmune": 2}
            sem_classes : Dict[str, int], optional
                Tissue type class dict. E.g. {"bg": 0, "epithel": 1, "stroma": 2}

        Example
        -------
            >>> # define type model, classes and paths
            >>> type_classes = {"bg": 0, "infl": 1, "neoplastic": 2}
            >>> model = cellpose_base(len(type_classes))
            >>> im_dir = "/path/to/ims"
            >>> lab_dir = "/path/to/labels/"

            >>> # Run inference first
            >>> inferer = ResizeInferer(
                    model=model,
                    input_path=im_dir,
                    checkpoint_path=ckpt_path,
                    out_activations={"type": "softmax", "cellpose": None},
                    out_boundary_weights={"type": None, "cellpose": None},
                    resize=(256, 256),
                    instance_postproc="cellpose",
                    normalization="minmax"
                )
            >>> inferer.infer()

            >>> bm = SegBenchmarker(
                    true_path=lab_dir,
                    inferer=inferer,
                    type_classes=type_classes
                )

            >>> # Run multi-class instance segmentation benchmarking
            >>> bm.run_inst_benchmark(how="multi")
        """
        if pred_path is None and inferer is None:
            raise ValueError(
                "Both `inferer` and `pred_path` cannot be set to None at the same time."
            )
        self.type_classes = type_classes
        self.sem_classes = sem_classes

        # solve the input paths, whether folder or h5 db.
        self.true_path = Path(true_path)
        if self.true_path.is_dir():
            self.true_ftype = "folder"
        elif self.true_path.is_file() and self.true_path.suffix in (".h5", ".hdf5"):
            self.true_ftype = "h5"
        else:
            raise ValueError(
                "The `true_path` needs to be a folder of .mat files or a h5 db."
            )

        if pred_path is not None:
            self.pred_path = Path(pred_path)
            if self.pred_path.is_dir():
                self.pred_ftype = "folder"
            elif self.pred_path.is_file() and self.pred_path.suffix in (".h5", ".hdf5"):
                self.pred_ftype = "h5"
            else:
                raise ValueError(
                    "The `true_path` needs to be a folder of .mat files or a h5 db."
                )
        else:
            self.pred_path = None
            self.pred_ftype = "inferer"
            try:
                inferer.out_masks
                inferer.soft_masks
                self.inferer = inferer
            except AttributeError:
                raise AttributeError(
                    "Did not find `out_masks` or `soft_masks` attributes. "
                    "To get these, run inference with `inferer.infer()`. "
                    "Remember to set `save_intermediate` to True for the inferer.`"
                )

        # define whether input files are from folder or h5-db
        if self.true_ftype == "folder":
            self.true_files = sorted(self.true_path.glob("*"))
            self.n_items = len(self.true_files)
        else:
            self.true_files = self.true_path
            self.n_items = FileHandler.read_h5_patch(
                self.true_files, 0, False, False, False, return_nitems=True
            )["nitems"]

        if self.pred_ftype == "folder":
            self.pred_files = sorted(self.pred_path.glob("*"))
        elif self.pred_ftype == "h5":
            self.pred_files = self.pred_path
        else:
            self.pred_files = None

    @staticmethod
    def compute_inst_metrics(
        true: np.ndarray,
        pred: np.ndarray,
        name: str,
        metrics: Tuple[str, ...] = ("pq",),
        type: str = "binary",
    ) -> Dict[str, float]:
        """Compute metrics for one prediciton-ground truth pair.

        Parameters
        ----------
            true : np.ndarray
                The ground truth instance labelled mask. Shape: (H, W).
            pred : np.ndarray
                The predicted instance labelled mask. Shape: (H, W).
            name : str
                Name of the sample/image.
            metrics : Tuple[str, ...], default=("pq", )
                A tuple of the metrics that will be computed. Allowed metrics:
                "pq", "aji", "dice2", "split_and_merge".

        Raises
        ------
            ValueError if an illegal metric is given.

        Returns
        -------
            Dict[str, float]:
                A dictionary where metric names are mapped to metric values.
                e.g. {"pq": 0.5, "aji": 0.55, "name": "sample1"}
        """
        allowed = list(INST_METRIC_LOOKUP.keys())
        if not all([m in allowed for m in metrics]):
            raise ValueError(
                f"An illegal metric was given. Got: {metrics}, allowed: {allowed}"
            )

        # Do not run metrics computation if there are no instances in neither of masks
        res = {}
        if len(np.unique(true)) > 1 or len(np.unique(pred)) > 1:
            true = remap_label(true)
            pred = remap_label(pred)

            met = {}
            for m in metrics:
                met[m] = INST_METRIC_LOOKUP[m]

            for k, m in met.items():
                score = m(true, pred)

                if isinstance(score, dict):
                    for n, v in score.items():
                        res[n] = v
                else:
                    res[k] = score

            res["name"] = name
            res["type"] = type
        else:
            res["name"] = name
            res["type"] = type

            for m in metrics:
                if m == "pq":
                    res["pq"] = -1.0
                    res["sq"] = -1.0
                    res["dq"] = -1.0
                else:
                    res[m] = -1.0

        return res

    @staticmethod
    def compute_sem_metrics(
        true: np.ndarray,
        pred: np.ndarray,
        name: str,
        num_classes: int,
        metrics: Tuple[str, ...] = ("iou",),
    ) -> Dict[str, float]:
        """Compute metrics for one prediciton-ground truth pair.

        Parameters
        ----------
            true : np.ndarray
                The ground truth semantic seg mask. Shape: (H, W).
            pred : np.ndarray
                The predicted semantic seg mask. Shape: (H, W).
            name : str
                Name of the sample/image.
            num_classes : int
                Number of classes in the dataset.
            metrics : Tuple[str, ...], default=("iou", )
                A tuple of the metrics that will be computed. Allowed metrics:
                "iou", "aji", "dice2", "split_and_merge".

        Raises
        ------
            ValueError if an illegal metric is given.

        Returns
        -------
            Dict[str, float]:
                A dictionary where metric names are mapped to metric values.
                e.g. {"iou": 0.5, "f1score": 0.55, "name": "sample1"}
        """
        if not isinstance(metrics, tuple) and not isinstance(metrics, list):
            raise ValueError("`metrics` must be either a list or tuple of values.")

        allowed = list(SEM_METRIC_LOOKUP.keys())
        if not all([m in allowed for m in metrics]):
            raise ValueError(
                f"An illegal metric was given. Got: {metrics}, allowed: {allowed}"
            )

        # Skip empty GTs
        met = {}
        for m in metrics:
            met[m] = SEM_METRIC_LOOKUP[m]

        res = {}
        for k, m in met.items():
            score = m(true, pred, num_classes=num_classes)
            res[k] = score

        res["name"] = name
        return res

    @staticmethod
    def run_metrics(
        args: List[Tuple[np.ndarray, np.ndarray, str, List[str]]],
        type: str = "inst",
    ) -> List[Dict[str, float]]:
        """Run segmentation metrics in parallel.

        Parameters
        ----------
            args : List[Tuple[np.ndarray, np.ndarray, str, List[str]]]
                A list of params for `compute_inst_metrics` or `compute_sem_metrics`
            type : str, default="inst"
                One of "inst" or "sem".

        Returns
        -------
            List[Dict[str, float]]:
                A list of dictionaries where metric names are mapped to metric values.
        """
        f = (
            SegBenchmarker._compute_sem_metrics
            if type == "sem"
            else SegBenchmarker._compute_inst_metrics
        )

        metrics = run_pool(f, args)

        return metrics

    def run_inst_benchmark(
        self, how: str, metrics: Tuple[str, ...] = ("pq",), chunk_size: int = 10
    ) -> None:
        """Run instance segmentation benchmarking.

        Parameters
        ----------
            how : str, default="binary"
                One of "binary" or "multi". "multi" computes metrics per class and
                "binary" computes metrics for all objects.
            metrics : Tuple[str, ...], default=("pq", )
                A tuple of the metrics that will be computed. Allowed metrics:
                "pq", "aji", "dice2" "ap".
            chunk_size : int, default=10
                Number of images in one chunk. Each chunk is send to multiprocessing.
                The benchmarking is by default run in chunks to not overflood the mem.

        Raises
        ------
            ValueError if an illegal `how` arg is given.

        Returns
        -------
            Dict[str, Any]:
                Dictionary mapping the metrics to values + metadata.
        """
        allowed = ("binary", "multi")
        if how not in allowed:
            raise ValueError(f"Illegal arg `how`. Got: {how}, Allowed: {allowed}")

        res = []
        if how == "multi" and self.type_classes is not None:
            for c, i in list(self.type_classes.items())[1:]:
                msg = c
                res.extend(
                    self._send_chunks_to_multiprocess(
                        metrics, "inst", msg, how=how, cls_val=i, chunk_size=chunk_size
                    )
                )
        else:
            msg = "binary"
            res.extend(
                self._send_chunks_to_multiprocess(
                    metrics, "inst", msg, how=how, chunk_size=chunk_size
                )
            )

        return res

    def run_sem_benchmark(
        self, metrics: Tuple[str, ...] = ("iou",), chunk_size: int = 10
    ) -> None:
        """Run semantic segmentation benchmarking.

        Parameters
        ----------
            metrics : Tuple[str, ...], default=("pq", )
                A tuple of the metrics that will be computed. Allowed metrics:
                "pq", "aji", "dice2" "ap".
            chunk_size : int, default=10
                Number of images in one chunk. Each chunk is send to multiprocessing.
                The benchmarking is by default run in chunks to not overflood the mem.


        Returns
        -------
            Dict[str, Any]:
                Dictionary mapping the metrics to values + metadata.
        """
        res = []
        msg = "semantic"
        res.extend(
            self._send_chunks_to_multiprocess(
                metrics, "sem", msg, chunk_size=chunk_size
            )
        )

        # re-format
        out = []
        for r in res:
            for k, val in self.sem_classes.items():
                datum = {"name": r["name"], "type": k}
                for m in metrics:
                    datum[m] = r[m][val]
                out.append(datum)

        return out

    def _read_one_mask_set(
        self,
        path: Union[Path, List[Path]],
        ix: int,
        ftype: str,
        return_inst: bool = True,
        return_type: bool = True,
        return_sem: bool = False,
        name: str = None,
    ) -> Dict[str, np.ndarray]:
        """Read a set of masks from a .mat file, h5-db or inferer object."""
        out = {}

        if ftype == "folder":
            name = path[ix].with_suffix("").name
            out["name"] = name

            masks = FileHandler.read_mat(path[ix], return_all=True)
            if return_inst:
                out["inst"] = masks["inst_map"]
            if return_type:
                out["type"] = masks["type_map"]
            if return_sem:
                out["sem"] = masks["sem_map"]
        elif ftype == "h5":
            masks = FileHandler.read_h5_patch(
                path=path,
                ix=ix,
                return_im=False,
                return_inst=return_inst,
                return_type=return_type,
                return_sem=return_sem,
                return_name=True,
            )
            out["name"] = masks["name"].with_suffix("").name

            if return_inst:
                out["inst"] = masks["inst"]
            if return_type:
                out["type"] = masks["type"]
            if return_sem:
                out["sem"] = masks["sem"]
        else:
            if name is None:
                name = path[ix].with_suffix("").name

            out["name"] = name
            masks = self.inferer.out_masks[name]
            if return_inst:
                out["inst"] = masks["inst"]
            if return_type:
                out["type"] = masks["type"]
            if return_sem:
                out["sem"] = masks["sem"]

        return out

    def _read_masks(
        self,
        true_path: Union[Path, List[Path]],
        pred_path: Union[Path, List[Path]],
        ix: int,
        return_inst: bool = True,
        return_type: bool = True,
        return_sem: bool = False,
    ) -> Tuple[Dict[str, np.ndarray], Dict[str, np.ndarray]]:
        """Read a set of pred & gt masks from either .mat files, h5 db or an inferer."""
        true_masks = self._read_one_mask_set(
            true_path,
            ix,
            self.true_ftype,
            return_inst=return_inst,
            return_type=return_type,
            return_sem=return_sem,
        )

        pred_masks = self._read_one_mask_set(
            pred_path if pred_path is not None else true_path,
            ix,
            self.pred_ftype,
            return_inst=return_inst,
            return_type=return_type,
            return_sem=return_sem,
            name=true_masks["name"] if self.pred_ftype == "inferer" else None,
        )

        return true_masks, pred_masks

    def _send_chunks_to_multiprocess(
        self,
        metrics: Tuple[str, ...],
        type: str,
        msg: str,
        how: str = None,
        chunk_size: int = 10,
        **kwargs,
    ) -> List[Dict[str, np.ndarray]]:
        """Send chunks one by one to multiprocessing."""
        res = []

        pbar_msg = f"Running {msg} {type} benchmark: "
        with tqdm(
            list(self._chunks(range(self.n_items), chunk_size)), desc=pbar_msg
        ) as pbar:
            for chunk in pbar:
                if type == "inst":
                    args = self._set_inst_args(
                        chunk, how=how, metrics=metrics, cls_name=msg, **kwargs
                    )
                else:
                    args = self._set_sem_args(chunk, metrics)

                met = self.run_metrics(args, type)
                res.extend([metric for metric in met if metric])

        return res

    def _set_inst_args(
        self,
        chunk: Iterable,
        how: str,
        metrics: Tuple[str, ...],
        cls_name: str = None,
        cls_val: int = None,
        **kwargs,
    ) -> List[Tuple[Any]]:
        """Generate arguments for instance seg benchmarking in parallel."""
        args = []
        for i in chunk:
            true_masks, pred_masks = self._read_masks(
                true_path=self.true_files,
                pred_path=self.pred_files,
                ix=i,
                return_inst=True,
                return_type=how == "multi",
            )

            # set args for inst benchmark
            name = true_masks["name"]
            if how == "multi":
                true_inst = true_masks["inst"]
                true_type = true_masks["type"]
                pred_inst = pred_masks["inst"]
                pred_type = pred_masks["type"]
                pred_type = get_type_instances(pred_inst, pred_type, cls_val)
                true_type = get_type_instances(true_inst, true_type, cls_val)
                arg_set = (true_type, pred_type, name, metrics, cls_name)
            else:
                true_inst = true_masks["inst"]
                pred_inst = pred_masks["inst"]
                arg_set = (true_inst, pred_inst, name, metrics)

            args.append(arg_set)

        return args

    def _set_sem_args(
        self,
        chunk: Iterable,
        metrics: Tuple[str, ...],
    ) -> List[Tuple[Any]]:
        """Generate arguments for semantic seg benchmarking in parallel."""
        args = []
        for i in chunk:
            true_masks, pred_masks = self._read_masks(
                true_path=self.true_files,
                pred_path=self.pred_files,
                ix=i,
                return_inst=False,
                return_type=False,
                return_sem=True,
            )
            # set args for inst benchmark
            name = true_masks["name"]
            args.append(
                (
                    true_masks["sem"],
                    pred_masks["sem"],
                    name,
                    len(self.sem_classes),
                    metrics,
                )
            )

        return args

    def _chunks(self, iterable: Iterable, size: int) -> Iterable:
        """Generate adjacent chunks of an iterable."""
        it = iter(iterable)
        return iter(lambda: tuple(itertools.islice(it, size)), ())

    def _compute_inst_metrics(
        args: List[Tuple[np.ndarray, np.ndarray, str, List[str]]]
    ) -> Dict[str, float]:
        return SegBenchmarker.compute_inst_metrics(*args)

    def _compute_sem_metrics(
        args: List[Tuple[np.ndarray, np.ndarray, str, List[str], int]]
    ) -> Dict[str, float]:
        return SegBenchmarker.compute_sem_metrics(*args)



import shutil
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import numpy as np

try:
    from ..datasets import SegmentationFolderDataset, SegmentationHDF5Dataset
    from ..datasets.dataset_writers.folder_writer import FolderWriter
    from ..datasets.dataset_writers.hdf5_writer import HDF5Writer
    from ._basemodule import BaseDataModule
    from .downloader import SimpleDownloader
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "To use the LizardDataModule, requests, pytorch-lightning, & albumentations "
        "libraries are needed. Install with "
        "`pip install requests pytorch-lightning albumentations`"
    )


class LizardDataModule(BaseDataModule):
    def __init__(
        self,
        save_dir: str,
        fold_split: Dict[str, int],
        img_transforms: List[str],
        inst_transforms: List[str],
        dataset_type: str = "folder",
        patch_size: Tuple[int, int] = (256, 256),
        stride: int = 128,
        normalization: str = None,
        batch_size: int = 8,
        num_workers: int = 8,
        **kwargs,
    ) -> None:
        """Set up Lizard datamodule. Creates overlapping patches of the Lizard dataset.

        The patches will be saved in directories:
        - `{save_dir}/train/*`
        - `{save_dir}/test/*`
        - `{save_dir}/valid/*`

        References
        ----------
        Graham, S., Jahanifar, M., Azam, A., Nimir, M., Tsang, Y.W., Dodd, K., Hero, E.,
        Sahota, H., Tank, A., Benes, K., & others (2021). Lizard: A Large-Scale Dataset
        for Colonic Nuclear Instance Segmentation and Classification. In Proceedings of
        the IEEE/CVF International Conference on Computer Vision (pp. 684-693).

        Parameters
        ----------
            save_dir : str
                Path to directory where the lizard data will be saved.
            fold_split : Dict[str, int]
                Defines how the folds are split into train, valid, and test sets.
                E.g. {"train": 1, "valid": 2, "test": 3}
            img_transforms : List[str]
                A list containing all the transformations that are applied to the input
                images and corresponding masks. Allowed ones: "blur", "non_spatial",
                "non_rigid", "rigid", "hue_sat", "random_crop", "center_crop", "resize"
            inst_transforms : List[str]
                A list containg all the transformations that are applied to only the
                instance labelled masks. Allowed ones: "cellpose", "contour", "dist",
                "edgeweight", "hovernet", "omnipose", "smooth_dist", "binarize"
            dataset_type : str, default="folder"
                The dataset type. One of "folder", "hdf5".
            patch_size : Tuple[int, int], default=(256, 256)
                The size of the patch extracted from the images.
            stride : int, default=128
                The stride of the sliding window patcher.
            normalization : str, optional
                Apply img normalization after all the transformations. One of "minmax",
                "norm", "percentile", None.
            batch_size : int, default=8
                Batch size for the dataloader.
            num_workers : int, default=8
                number of cpu cores/threads used in the dataloading process.

        Example
        -------
            >>> from pathlib import Path
            >>> from cellseg_models_pytorch.datamodules import LizardDataModule

            >>> fold_split = {"train": 1, "valid": 2, "test": 3}
            >>> save_dir = Path.home() / "lizard"
            >>> lizard_module = LizardDataModule(
                    save_dir=save_dir,
                    fold_split=fold_split,
                    inst_transforms=["dist", "stardist"],
                    img_transforms=["blur", "hue_sat"],
                    normalization="percentile",
                    dataset_type="hdf5",
                    patch_size=(320, 320),
                    stride=128
                )

            >>> # lizard_module.download(save_dir) # just the downloading
            >>> lizard_module.prepare_data(tiling=True) # downloading & processing
        """
        super().__init__(batch_size, num_workers)
        self.save_dir = Path(save_dir)
        self.fold_split = fold_split
        self.patch_size = patch_size
        self.stride = stride
        self.img_transforms = img_transforms
        self.inst_transforms = inst_transforms
        self.normalization = normalization
        self.kwargs = kwargs if kwargs is not None else {}

        if dataset_type not in ("folder", "hdf5"):
            raise ValueError(
                f"Illegal `dataset_type` arg. Got {dataset_type}. "
                f"Allowed: {('folder', 'hdf5')}"
            )

        self.dataset_type = dataset_type

    @property
    def type_classes(self) -> Dict[str, int]:
        """Lizard cell type classes."""
        return {
            "bg": 0,
            "neutrophil": 1,
            "epithelial": 2,
            "lymphocyte": 3,
            "plasma": 4,
            "eosinophil": 5,
            "connective": 6,
        }

    @staticmethod
    def download(root: str) -> None:
        """Download the lizard dataset from online."""
        for ix in [1, 2]:
            fn = f"lizard_images{ix}.zip"
            url = f"https://warwick.ac.uk/fac/cross_fac/tia/data/lizard/{fn}"
            SimpleDownloader.download(url, root)

        url = "https://warwick.ac.uk/fac/cross_fac/tia/data/lizard/lizard_labels.zip"
        SimpleDownloader.download(url, root)
        LizardDataModule.extract_zips(root, rm=True)

    def prepare_data(self, rm_orig: bool = False, tiling: bool = True) -> None:
        """Prepare the lizard datasets.

        1. Download lizard folds from:
            "https://warwick.ac.uk/fac/cross_fac/tia/data/lizard/"
        2. split the images and masks into train, valid and test sets.
        3. Patch the images such that they can be used with the datasets/loaders.

        Parameters
        ----------
            rm_orig : bool, default=False
                After processing all the files, If True, removes the original
                un-processed files.
            tiling : bool, default=True
                Flag, whether to cut images into tiles. Can be set to False if you only
                want to download and split the data and then work it out on your own.
        """
        folders_found = [
            d.name
            for d in self.save_dir.iterdir()
            if d.name.lower() in ("lizard_images1", "lizard_images2", "lizard_labels")
            and d.is_dir()
        ]
        phases_found = [
            d.name
            for d in self.save_dir.iterdir()
            if d.name in ("train", "test", "valid") and d.is_dir()
        ]

        patches_found = []
        if phases_found:
            patches_found = [
                sub_d.name
                for d in self.save_dir.iterdir()
                if d.name in ("train", "test", "valid") and d.is_dir()
                for sub_d in d.iterdir()
                if sub_d.name
                in (
                    f"{d.name}_im_patches",
                    f"{d.name}_mask_patches",
                    f"{d.name}_patches",
                )
                and any(sub_d.iterdir())
            ]

        if len(folders_found) < 3 and not phases_found:
            print(
                "Found no data or an incomplete dataset. Downloading the whole thing..."
            )
            for d in self.save_dir.iterdir():
                shutil.rmtree(d)
            LizardDataModule.download(self.save_dir)
        else:
            print("Found all folds. Skip downloading.")

        if not phases_found:
            print("Splitting the files into train, valid, and test sets.")
            for phase, fold_ix in self.fold_split.items():
                img_dir1 = self.save_dir / "Lizard_Images1"
                img_dir2 = self.save_dir / "Lizard_Images2"
                label_dir = self.save_dir / "Lizard_Labels"
                save_im_dir = self.save_dir / phase / "images"
                save_mask_dir = self.save_dir / phase / "labels"

                self._split_to_fold(
                    img_dir1,
                    img_dir2,
                    label_dir,
                    save_im_dir,
                    save_mask_dir,
                    fold_ix,
                    not rm_orig,
                )
        else:
            print(
                "Found splitted Lizard data. "
                "If in need of a re-download, please empty the `save_dir` folder."
            )

        if rm_orig:
            for d in self.save_dir.iterdir():
                if "lizard" in d.name.lower() or "macosx" in d.name.lower():
                    shutil.rmtree(d)

        if tiling and not patches_found:
            print("Patch the data... This will take a while...")
            for phase in self.fold_split.keys():
                save_im_dir = self.save_dir / phase / "images"
                save_mask_dir = self.save_dir / phase / "labels"

                if self.dataset_type == "hdf5":
                    sdir = self.save_dir / phase / f"{phase}_patches"
                    sdir.mkdir(parents=True, exist_ok=True)
                    writer = HDF5Writer(
                        in_dir_im=save_im_dir,
                        in_dir_mask=save_mask_dir,
                        save_dir=sdir,
                        file_name=f"lizard_{phase}.h5",
                        patch_size=self.patch_size,
                        stride=self.stride,
                        transforms=["rigid"],
                    )
                else:
                    sdir_im = self.save_dir / phase / f"{phase}_im_patches"
                    sdir_mask = self.save_dir / phase / f"{phase}_mask_patches"
                    sdir_im.mkdir(parents=True, exist_ok=True)
                    sdir_mask.mkdir(parents=True, exist_ok=True)
                    writer = FolderWriter(
                        in_dir_im=save_im_dir,
                        in_dir_mask=save_mask_dir,
                        save_dir_im=sdir_im,
                        save_dir_mask=sdir_mask,
                        patch_size=self.patch_size,
                        stride=self.stride,
                        transforms=["rigid"],
                    )
                writer.write(tiling=True, pre_proc=self._process_label, msg=phase)
        else:
            print(
                "Found processed Lizard data. "
                "If in need of a re-process, please empty the `save_dir` folders."
            )

    def _get_path(self, phase: str, dstype: str, is_mask: bool = False) -> Path:
        if dstype == "hdf5":
            p = self.save_dir / phase / f"{phase}_patches" / f"lizard_{phase}.h5"
        else:
            dtype = "mask" if is_mask else "im"
            p = self.save_dir / phase / f"{phase}_{dtype}_patches"

        return p

    def setup(self, stage: Optional[str] = None) -> None:
        """Set up the train, valid, and test datasets."""
        if self.dataset_type == "hdf5":
            DS = SegmentationHDF5Dataset
        else:
            DS = SegmentationFolderDataset

        self.trainset = DS(
            path=self._get_path("train", self.dataset_type, is_mask=False),
            mask_path=self._get_path("train", self.dataset_type, is_mask=True),
            img_transforms=self.img_transforms,
            inst_transforms=self.inst_transforms,
            return_sem=False,
            normalization=self.normalization,
            **self.kwargs,
        )

        self.validset = DS(
            path=self._get_path("valid", self.dataset_type, is_mask=False),
            mask_path=self._get_path("valid", self.dataset_type, is_mask=True),
            img_transforms=self.img_transforms,
            inst_transforms=self.inst_transforms,
            return_sem=False,
            normalization=self.normalization,
            **self.kwargs,
        )

        self.testset = DS(
            path=self._get_path("test", self.dataset_type, is_mask=False),
            mask_path=self._get_path("test", self.dataset_type, is_mask=True),
            img_transforms=self.img_transforms,
            inst_transforms=self.inst_transforms,
            return_sem=False,
            normalization=self.normalization,
            **self.kwargs,
        )

    def _split_to_fold(
        self,
        img_dir1: Path,
        img_dir2: Path,
        label_dir: Path,
        save_im_dir: Path,
        save_mask_dir: Path,
        fold: int,
        copy: bool = True,
    ) -> None:
        """Move the downloaded data split into one of 'train', 'valid' or 'test' dir."""
        Path(save_im_dir).mkdir(parents=True, exist_ok=True)
        Path(save_mask_dir).mkdir(parents=True, exist_ok=True)
        info_path = label_dir / "info.csv"
        info = np.genfromtxt(info_path, dtype="str", delimiter=",", skip_header=True)
        info = info[info[:, -1] == str(fold)]
        for i in range(info.shape[0]):
            fn, _, _ = info[i]

            p1 = img_dir1 / f"{fn}.png"
            p2 = img_dir2 / f"{fn}.png"
            src_im = p1 if p1.exists() else p2
            src_mask = label_dir / "Labels" / f"{fn}.mat"

            if copy:
                shutil.copy(src_im, save_im_dir)
                shutil.copy(src_mask, save_mask_dir)
            else:
                src_im.rename(save_im_dir / src_im.name)
                src_mask.rename(save_mask_dir / src_mask.name)

    def _process_label(self, label: np.ndarray) -> None:
        """Process the labels.

        NOTE: this is done to match the schema that's used by the Dataset classes.
        """
        inst_map = label["inst_map"]
        classes = label["class"]
        nuclei_id = label["id"]

        type_map = np.zeros_like(inst_map)
        unique_values = np.unique(inst_map).tolist()[1:]  # remove 0
        nuclei_id = np.squeeze(nuclei_id).tolist()
        for value in unique_values:
            # Get the position of the corresponding value
            inst = np.copy(inst_map == value)
            idx = nuclei_id.index(value)

            class_ = classes[idx]
            type_map[inst > 0] = class_

        return {"inst_map": inst_map, "type_map": type_map}



from typing import Any, Callable, Optional

import torch

from ..functional.train_metrics import accuracy, iou

try:
    from torchmetrics import (
        MeanSquaredError,
        Metric,
        StructuralSimilarityIndexMeasure,
        UniversalImageQualityIndex,
    )

except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "`torchmetrics` package is required when using metric callbacks. "
        "Install with `pip install torchmetrics`"
    )


__all__ = ["Accuracy", "MeanIoU"]


class Accuracy(Metric):
    higher_is_better: Optional[bool] = True
    full_state_update: bool = False

    def __init__(
        self,
        compute_on_step: bool = True,
        dist_sync_on_step: bool = False,
        progress_group: Any = None,
        dist_sync_func: Callable = None,
        **kwargs
    ) -> None:
        """Create a custom torchmetrics accuracy callback.

        Parameters
        ----------
            compute_on_step : bool, default=True
                Forward only calls update() and returns None if this is set to False.
            dist_sync_on_step : bool, default=False
                Synchronize computed values in distributed setting
            process_group : any, optional
                Specify the process group on which synchronization is called.
                default: None (which selects the entire world)
            dist_sync_func : Callable, optional
                Callback that performs the allgather operation on the metric state.
                When None, DDP will be used to perform the allgather.
        """
        super().__init__(
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=progress_group,
            dist_sync_fn=dist_sync_func,
        )

        self.add_state(
            "batch_accuracies", default=torch.tensor(0.0), dist_reduce_fx="sum"
        )

        self.add_state("n_batches", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(
        self,
        preds: torch.Tensor,
        target: torch.Tensor,
        activation: str = "softmax",
        **kwargs
    ) -> None:
        """Update the batch accuracy list with one batch accuracy value.

        Parameters
        ----------
            pred : torch.Tensor
                Predicted output from the model. Shape (B, C, H, W).
            target : torch.Tensor
                The ground truth segmentation tensor. Shape (B, H, W).
            activation : str, default="softmax"
                The activation function. One of: "softmax", "sigmoid" or None.
        """
        batch_acc = accuracy(preds, target, activation)
        self.batch_accuracies += batch_acc
        self.n_batches += 1

    def compute(self) -> torch.Tensor:
        """Compute the accuracy of one batch and normalize accordingly.

        Returns
        -------
            torch.Tensor: The accuracy value. Shape (1).
        """
        return self.batch_accuracies / self.n_batches


class MeanIoU(Metric):
    higher_is_better: Optional[bool] = True
    full_state_update: bool = False

    def __init__(
        self,
        compute_on_step: bool = True,
        dist_sync_on_step: bool = False,
        progress_grouo: Any = None,
        dist_sync_func: Callable = None,
        **kwargs
    ) -> None:
        """Create a custom torchmetrics mIoU callback.

        Parameters
        ----------
            compute_on_step : bool, default=True
                Forward only calls update() and returns None if this is set to False.
            dist_sync_on_step : bool, default=False
                Synchronize computed values in distributed setting
            process_group : any, optional
                Specify the process group on which synchronization is called.
                default: None (which selects the entire world)
            dist_sync_func : Callable, optional
                Callback that performs the allgather operation on the metric state.
                When None, DDP will be used to perform the allgather.
        """
        super().__init__(
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=progress_grouo,
            dist_sync_fn=dist_sync_func,
        )

        self.add_state("batch_ious", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("n_batches", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(
        self,
        preds: torch.Tensor,
        target: torch.Tensor,
        activation: str = "softmax",
        **kwargs
    ) -> None:
        """Update the batch IoU list with one batch IoU matrix.

        Parameters
        ----------
            pred : torch.Tensor
                Predicted output from the model. Shape (B, C, H, W).
            target : torch.Tensor
                The ground truth segmentation tensor. Shape (B, H, W).
            activation : str, default="softmax"
                The activation function. One of: "softmax", "sigmoid" or None.
        """
        batch_iou = iou(preds, target, activation)
        self.batch_ious += batch_iou.mean()
        self.n_batches += 1

    def compute(self) -> torch.Tensor:
        """Normalize the batch IoU values.

        Returns
        -------
            torch.Tensor: The IoU mat. Shape (B, n_classes, n_classes).
        """
        return self.batch_ious / self.n_batches


METRIC_LOOKUP = {
    "acc": Accuracy,
    "miou": MeanIoU,
    "mse": MeanSquaredError,
    "ssim": StructuralSimilarityIndexMeasure,
    "iqi": UniversalImageQualityIndex,
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cellseg_models_pytorch',
 'cellseg_models_pytorch.datamodules',
 'cellseg_models_pytorch.datamodules.tests',
 'cellseg_models_pytorch.datasets',
 'cellseg_models_pytorch.datasets.dataset_writers',
 'cellseg_models_pytorch.datasets.tests',
 'cellseg_models_pytorch.decoders',
 'cellseg_models_pytorch.decoders.long_skips',
 'cellseg_models_pytorch.decoders.long_skips.merging',
 'cellseg_models_pytorch.decoders.tests',
 'cellseg_models_pytorch.inference',
 'cellseg_models_pytorch.inference.tests',
 'cellseg_models_pytorch.losses',
 'cellseg_models_pytorch.losses.criterions',
 'cellseg_models_pytorch.losses.tests',
 'cellseg_models_pytorch.metrics',
 'cellseg_models_pytorch.metrics.tests',
 'cellseg_models_pytorch.models',
 'cellseg_models_pytorch.models.base',
 'cellseg_models_pytorch.models.cellpose',
 'cellseg_models_pytorch.models.hovernet',
 'cellseg_models_pytorch.models.stardist',
 'cellseg_models_pytorch.models.tests',
 'cellseg_models_pytorch.modules',
 'cellseg_models_pytorch.modules.act',
 'cellseg_models_pytorch.modules.conv',
 'cellseg_models_pytorch.modules.norm',
 'cellseg_models_pytorch.modules.self_attention',
 'cellseg_models_pytorch.modules.tests',
 'cellseg_models_pytorch.modules.upsample',
 'cellseg_models_pytorch.optimizers',
 'cellseg_models_pytorch.optimizers.tests',
 'cellseg_models_pytorch.postproc',
 'cellseg_models_pytorch.postproc.functional',
 'cellseg_models_pytorch.postproc.functional.cellpose',
 'cellseg_models_pytorch.postproc.functional.stardist',
 'cellseg_models_pytorch.postproc.tests',
 'cellseg_models_pytorch.training',
 'cellseg_models_pytorch.training.callbacks',
 'cellseg_models_pytorch.training.functional',
 'cellseg_models_pytorch.training.lit',
 'cellseg_models_pytorch.training.tests',
 'cellseg_models_pytorch.transforms',
 'cellseg_models_pytorch.transforms.albu_transforms',
 'cellseg_models_pytorch.transforms.functional',
 'cellseg_models_pytorch.transforms.tests',
 'cellseg_models_pytorch.utils',
 'cellseg_models_pytorch.utils.tests']

package_data = \
{'': ['*'],
 'cellseg_models_pytorch.datasets.tests': ['data/*',
                                           'data/imgs/*',
                                           'data/masks/*'],
 'cellseg_models_pytorch.inference.tests': ['data/*'],
 'cellseg_models_pytorch.training.tests': ['data/*'],
 'cellseg_models_pytorch.utils.tests': ['data/*']}

install_requires = \
['numba>=0.55.2,<0.58.0',
 'numpy>=1.22,<1.24',
 'opencv-python>=4.2.0.32,<5.0.0.0',
 'pathos>=0.2.8,<0.4.0',
 'scikit-image>=0.19,<0.20',
 'scikit-learn>=1.0.2,<2.0.0',
 'scipy>=1.7,<2.0',
 'timm>=0.5.4,<0.7.0',
 'torch>=1.8.1,<2.0.0',
 'tqdm>=4.64.0,<5.0.0']

extras_require = \
{'all': ['pytorch-lightning>=1.6.0,<2.0.0',
         'tables>=3.6.0,<4.0.0',
         'albumentations>=1.0.0,<2.0.0',
         'requests>=2.28.0,<3.0.0',
         'geojson>=2.5.0,<3.0.0',
         'torchmetrics>=0.10,<0.12']}

setup_kwargs = {
    'name': 'cellseg-models-pytorch',
    'version': '0.1.20',
    'description': 'Python library for 2D cell/nuclei instance segmentation models written with PyTorch.',
    'long_description': '<div align="center">\n\n![Logo](./images/logo.png)\n\n**Python library for 2D cell/nuclei instance segmentation models written with [PyTorch](https://pytorch.org/).**\n\n[![Generic badge](https://img.shields.io/badge/License-MIT-<COLOR>.svg?style=for-the-badge)](https://github.com/okunator/cellseg_models.pytorch/blob/master/LICENSE)\n[![PyTorch - Version](https://img.shields.io/badge/PYTORCH-1.8.1+-red?style=for-the-badge&logo=pytorch)](https://pytorch.org/)\n[![Python - Version](https://img.shields.io/badge/PYTHON-3.8+-red?style=for-the-badge&logo=python&logoColor=white)](https://www.python.org/)\n<br>\n[![Github Test](https://img.shields.io/github/actions/workflow/status/okunator/cellseg_models.pytorch/tests.yml?label=Tests&logo=github&&style=for-the-badge)](https://github.com/okunator/cellseg_models.pytorch/actions/workflows/tests.yml)\n[![Pypi](https://img.shields.io/pypi/v/cellseg-models-pytorch?color=blue&logo=pypi&style=for-the-badge)](https://pypi.org/project/cellseg-models-pytorch/)\n[![Codecov](https://img.shields.io/codecov/c/github/okunator/cellseg_models.pytorch?logo=codecov&style=for-the-badge&token=oGSj7FZ1lm)](https://codecov.io/gh/okunator/cellseg_models.pytorch)\n<br>\n[![DOI](https://zenodo.org/badge/450787123.svg)](https://zenodo.org/badge/latestdoi/450787123)\n\n</div>\n\n<div align="center">\n\n</div>\n\n## Introduction\n\n**cellseg-models.pytorch** is a library built upon [PyTorch](https://pytorch.org/) that contains multi-task encoder-decoder architectures along with dedicated post-processing methods for segmenting cell/nuclei instances. As the name might suggest, this library is heavily inspired by [segmentation_models.pytorch](https://github.com/qubvel/segmentation_models.pytorch) library for semantic segmentation.\n\n## Features\n\n- High level API to define cell/nuclei instance segmentation models.\n- 4 cell/nuclei instance segmentation models and more to come.\n- Open source datasets for training and benchmarking.\n- Pre-trained backbones/encoders from the [timm](https://github.com/rwightman/pytorch-image-models) library.\n- All the architectures can be augmented to **panoptic segmentation**.\n- A lot of flexibility to modify the components of the model architectures.\n- Sliding window inference for large images.\n- Multi-GPU inference.\n- Popular training losses and benchmarking metrics.\n- Simple model training with [pytorch-lightning](https://www.pytorchlightning.ai/).\n- Benchmarking utilities both for model latency & segmentation performance.\n- Regularization techniques to tackle batch effects/domain shifts.\n- Ability to add transformers to the decoder layers.\n\n## Installation\n\n**Basic installation**\n\n```shell\npip install cellseg-models-pytorch\n```\n\n**To install extra dependencies (training utilities and datamodules for open-source datasets) use**\n\n```shell\npip install cellseg-models-pytorch[all]\n```\n\n## Models\n\n| Model                      | Paper                                                                          |\n| -------------------------- | ------------------------------------------------------------------------------ |\n| [[1](#Citation)] HoVer-Net | https://www.sciencedirect.com/science/article/pii/S1361841519301045?via%3Dihub |\n| [[2](#Citation)] Cellpose  | https://www.nature.com/articles/s41592-020-01018-x                             |\n| [[3](#Citation)] Omnipose  | https://www.biorxiv.org/content/10.1101/2021.11.03.467199v2                    |\n| [[4](#Citation)] Stardist  | https://arxiv.org/abs/1806.03535                                               |\n\n## Datasets\n\n| Dataset                       | Paper                                                                                            |\n| ----------------------------- | ------------------------------------------------------------------------------------------------ |\n| [[5, 6](#References)] Pannuke | https://arxiv.org/abs/2003.10778 , https://link.springer.com/chapter/10.1007/978-3-030-23937-4_2 |\n| [[7](#References)] Lizard     | http://arxiv.org/abs/2108.11195                                                                  |\n\n## Notebook examples\n\n- [Training Stardist with Pannuke](https://github.com/okunator/cellseg_models.pytorch/blob/main/examples/pannuke_nuclei_segmentation_stardist.ipynb). Train the Stardist model with constant sized Pannuke patches.\n- [Training Cellpose with Lizard](https://github.com/okunator/cellseg_models.pytorch/blob/main/examples/lizard_nuclei_segmentation_cellpose.ipynb). Train the Cellpose model with Lizard dataset that is composed of varying sized images.\n- [Benchmarking Cellpose Trained on Pannuke](https://github.com/okunator/cellseg_models.pytorch/blob/main/examples/pannuke_cellpose_benchmark.ipynb). Benchmark Cellpose trained on Pannuke. Both the model performance and latency.\n\n## Code Examples\n\n**Define Cellpose for cell segmentation.**\n\n```python\nimport cellseg_models_pytorch as csmp\nimport torch\n\nmodel = csmp.models.cellpose_base(type_classes=5)\nx = torch.rand([1, 3, 256, 256])\n\n# NOTE: the outputs still need post-processing.\ny = model(x) # {"cellpose": [1, 2, 256, 256], "type": [1, 5, 256, 256]}\n```\n\n**Define Cellpose for cell and tissue area segmentation (Panoptic segmentation).**\n\n```python\nimport cellseg_models_pytorch as csmp\nimport torch\n\nmodel = csmp.models.cellpose_plus(type_classes=5, sem_classes=3)\nx = torch.rand([1, 3, 256, 256])\n\n# NOTE: the outputs still need post-processing.\ny = model(x) # {"cellpose": [1, 2, 256, 256], "type": [1, 5, 256, 256], "sem": [1, 3, 256, 256]}\n```\n\n**Define panoptic Cellpose model with more flexibility.**\n\n```python\nimport cellseg_models_pytorch as csmp\n\n# the model will include two decoder branches.\ndecoders = ("cellpose", "sem")\n\n# and in total three segmentation heads emerging from the decoders.\nheads = {\n    "cellpose": {"cellpose": 2, "type": 5},\n    "sem": {"sem": 3}\n}\n\nmodel = csmp.CellPoseUnet(\n    decoders=decoders,                   # cellpose and semantic decoders\n    heads=heads,                         # three output heads\n    depth=5,                             # encoder depth\n    out_channels=(256, 128, 64, 32, 16), # num out channels at each decoder stage\n    layer_depths=(4, 4, 4, 4, 4),        # num of conv blocks at each decoder layer\n    style_channels=256,                  # num of style vector channels\n    enc_name="resnet50",                 # timm encoder\n    enc_pretrain=True,                   # imagenet pretrained encoder\n    long_skip="unetpp",                  # unet++ long skips ("unet", "unetpp", "unet3p")\n    merge_policy="sum",                  # concatenate long skips ("cat", "sum")\n    short_skip="residual",               # residual short skips ("basic", "residual", "dense")\n    normalization="bcn",                 # batch-channel-normalization.\n    activation="gelu",                   # gelu activation.\n    convolution="wsconv",                # weight standardized conv.\n    attention="se",                      # squeeze-and-excitation attention.\n    pre_activate=False,                  # normalize and activation after convolution.\n)\n\nx = torch.rand([1, 3, 256, 256])\n\n# NOTE: the outputs still need post-processing.\ny = model(x) # {"cellpose": [1, 2, 256, 256], "type": [1, 5, 256, 256], "sem": [1, 3, 256, 256]}\n```\n\n**Run HoVer-Net inference and post-processing with a sliding window approach.**\n\n```python\nimport cellseg_models_pytorch as csmp\n\n# define the model\nmodel = csmp.models.hovernet_base(type_classes=5)\n\n# define the final activations for each model output\nout_activations = {"hovernet": "tanh", "type": "softmax", "inst": "softmax"}\n\n# define whether to weight down the predictions at the image boundaries\n# typically, models perform the poorest at the image boundaries and with\n# overlapping patches this causes issues which can be overcome by down-\n# weighting the prediction boundaries\nout_boundary_weights = {"hovernet": True, "type": False, "inst": False}\n\n# define the inferer\ninferer = csmp.inference.SlidingWindowInferer(\n    model=model,\n    input_folder="/path/to/images/",\n    checkpoint_path="/path/to/model/weights/",\n    out_activations=out_activations,\n    out_boundary_weights=out_boundary_weights,\n    instance_postproc="hovernet",               # THE POST-PROCESSING METHOD\n    normalization="percentile",                 # same normalization as in training\n    patch_size=(256, 256),\n    stride=128,\n    padding=80,\n    batch_size=8,\n)\n\ninferer.infer()\n\ninferer.out_masks\n# {"image1" :{"inst": [H, W], "type": [H, W]}, ..., "imageN" :{"inst": [H, W], "type": [H, W]}}\n```\n\n## Models API\n\nGenerally, the model building API enables the effortless creation of hard-parameter sharing multi-task encoder-decoder CNN architectures. The general architectural schema is illustrated in the below image.\n\n<br><br>\n![Architecture](./images/architecture_overview.png)\n\n### Class API\n\nThe class API enables the most flexibility in defining different model architectures. It borrows a lot from [segmentation_models.pytorch](https://github.com/qubvel/segmentation_models.pytorch) models API.\n\n**Model classes**:\n\n- `csmp.CellPoseUnet`\n- `csmp.StarDistUnet`\n- `csmp.HoverNet`\n\n**All of the models contain**:\n\n- `model.encoder` - pretrained [timm](https://github.com/rwightman/pytorch-image-models) backbone for feature extraction.\n- `model.{decoder_name}_decoder` - Models can have multiple decoders with unique names.\n- `model.{head_name}_seg_head` - Model decoders can have multiple segmentation heads with unique names.\n- `model.forward(x)` - forward pass.\n- `model.forward_features(x)` - forward pass of the encoder and decoders. Returns enc and dec features\n\n**Defining your own multi-task architecture**\n\nFor example, to define a multi-task architecture that has `resnet50` encoder, four decoders, and 5 output heads with `CellPoseUnet` architectural components, we could do this:\n\n```python\nimport cellseg_models_pytorch as csmp\nimport torch\n\nmodel = csmp.CellPoseUnet(\n    decoders=("cellpose", "dist", "contour", "sem"),\n    heads={\n        "cellpose": {"type": 5, "cellpose": 2},\n        "dist": {"dist": 1},\n        "contour": {"contour": 1},\n        "sem": {"sem": 4}\n    },\n)\n\nx = torch.rand([1, 3, 256, 256])\nmodel(x)\n# {\n#   "cellpose": [1, 2, 256, 256],\n#   "type": [1, 5, 256, 256],\n#   "dist": [1, 1, 256, 256],\n#   "contour": [1, 1, 256, 256],\n#   "sem": [1, 4, 256, 256]\n# }\n```\n\n### Function API\n\nWith the function API, you can build models with low effort by calling the below listed functions. Under the hood, the function API simply calls the above classes with pre-defined decoder and head names. The training and post-processing tools of this library are built around these names, thus, it is recommended to use the function API, although, it is a bit more rigid than the class API. Basically, the function API only lacks the ability to define the output-tasks of the model, but allows for all the rest as the class API.\n\n| Model functions                        | Output names                              | Task                             |\n| -------------------------------------- | ----------------------------------------- | -------------------------------- |\n| `csmp.models.cellpose_base`            | `"type"`, `"cellpose"`,                   | **instance segmentation**        |\n| `csmp.models.cellpose_plus`            | `"type"`, `"cellpose"`, `"sem"`,          | **panoptic segmentation**        |\n| `csmp.models.omnipose_base`            | `"type"`, `"omnipose"`                    | **instance segmentation**        |\n| `csmp.models.omnipose_plus`            | `"type"`, `"omnipose"`, `"sem"`,          | **panoptic segmentation**        |\n| `csmp.models.hovernet_base`            | `"type"`, `"inst"`, `"hovernet"`          | **instance segmentation**        |\n| `csmp.models.hovernet_plus`            | `"type"`, `"inst"`, `"hovernet"`, `"sem"` | **panoptic segmentation**        |\n| `csmp.models.hovernet_small`           | `"type"`,`"hovernet"`                     | **instance segmentation**        |\n| `csmp.models.hovernet_small_plus`      | `"type"`, `"hovernet"`, `"sem"`           | **panoptic segmentation**        |\n| `csmp.models.stardist_base`            | `"stardist"`, `"dist"`                    | **binary instance segmentation** |\n| `csmp.models.stardist_base_multiclass` | `"stardist"`, `"dist"`, `"type"`          | **instance segmentation**        |\n| `csmp.models.stardist_plus`            | `"stardist"`, `"dist"`, `"type"`, `"sem"` | **panoptic segmentation**        |\n\n## References\n\n- [1] S. Graham, Q. D. Vu, S. E. A. Raza, A. Azam, Y-W. Tsang, J. T. Kwak and N. Rajpoot. "HoVer-Net: Simultaneous Segmentation and Classification of Nuclei in Multi-Tissue Histology Images." Medical Image Analysis, Sept. 2019.\n- [2] Stringer, C.; Wang, T.; Michaelos, M. & Pachitariu, M. Cellpose: a generalist algorithm for cellular segmentation Nature Methods, 2021, 18, 100-106\n- [3] Cutler, K. J., Stringer, C., Wiggins, P. A., & Mougous, J. D. (2022). Omnipose: a high-precision morphology-independent solution for bacterial cell segmentation. bioRxiv. doi:10.1101/2021.11.03.467199\n- [4] Uwe Schmidt, Martin Weigert, Coleman Broaddus, & Gene Myers (2018). Cell Detection with Star-Convex Polygons. In Medical Image Computing and Computer Assisted Intervention - MICCAI 2018 - 21st International Conference, Granada, Spain, September 16-20, 2018, Proceedings, Part II (pp. 265–273).\n- [5] Gamper, J., Koohbanani, N., Benet, K., Khuram, A., & Rajpoot, N. (2019) PanNuke: an open pan-cancer histology dataset for nuclei instance segmentation and classification. In European Congress on Digital Pathology (pp. 11-19).\n- [6] Gamper, J., Koohbanani, N., Graham, S., Jahanifar, M., Khurram, S., Azam, A.,Hewitt, K., & Rajpoot, N. (2020). PanNuke Dataset Extension, Insights and Baselines. arXiv preprint arXiv:2003.10778.\n- [7] Graham, S., Jahanifar, M., Azam, A., Nimir, M., Tsang, Y.W., Dodd, K., Hero, E., Sahota, H., Tank, A., Benes, K., & others (2021). Lizard: A Large-Scale Dataset for Colonic Nuclear Instance Segmentation and Classification. In Proceedings of the IEEE/CVF International Conference on Computer Vision (pp. 684-693).\n\n## Citation\n\n```bibtex\n@misc{csmp2022,\n    title={{cellseg_models.pytorch}: Cell/Nuclei Segmentation Models and Benchmark.},\n    author={Oskari Lehtonen},\n    howpublished = {\\url{https://github.com/okunator/cellseg_models.pytorch}},\n    doi = {10.5281/zenodo.7064617}\n    year={2022}\n}\n```\n\n## Licence\n\nThis project is distributed under [MIT License](https://github.com/okunator/cellseg_models.pytorch/blob/main/LICENSE)\n\nThe project contains code from the original cell segmentation and 3rd-party libraries that have permissive licenses:\n\n- [timm](https://github.com/rwightman/pytorch-image-models) (Apache-2)\n- [HoVer-Net](https://github.com/vqdang/hover_net) (MIT)\n- [Cellpose](https://github.com/MouseLand/cellpose) (BSD-3)\n- [Stardist](https://github.com/stardist/stardist) (BSD-3)\n\nIf you find this library useful in your project, it is your responsibility to ensure you comply with the conditions of any dependent licenses. Please create an issue if you think something is missing regarding the licenses.\n',
    'author': 'Okunator',
    'author_email': 'oskari.lehtonen@helsinki.fi',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/okunator/cellseg_models.pytorch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



from typing import List

import timm
import torch
import torch.nn as nn

__all__ = ["TimmEncoder"]


class TimmEncoder(nn.Module):
    def __init__(
        self,
        name: str,
        pretrained: bool = True,
        in_channels: int = 3,
        depth: int = 5,
        **kwargs
    ) -> None:
        """Import any encoder from timm package."""
        super().__init__()

        kwargs = dict(
            in_chans=in_channels,
            features_only=True,
            pretrained=pretrained,
            out_indices=tuple(range(depth)),
        )

        self.model = timm.create_model(name, **kwargs)

        self.in_channels = in_channels
        self.out_channels = tuple(self.model.feature_info.channels()[::-1])

    def forward(self, x: torch.Tensor) -> List[torch.Tensor]:
        """Forward pass of the encoder and return all the features."""
        features = self.model(x)
        return features[::-1]



from typing import Any, Dict, Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from .base_modules import Identity
from .misc_modules import LayerScale
from .mlp import MlpBlock
from .patch_embeddings import ContiguousEmbed
from .self_attention_modules import SelfAttentionBlock

__all__ = ["Transformer2D", "TransformerLayer"]


class Transformer2D(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_heads: int = 8,
        head_dim: int = 64,
        cross_attention_dim: int = None,
        n_blocks: int = 2,
        block_types: Tuple[str, ...] = ("exact", "exact"),
        computation_types: Tuple[str, ...] = ("basic", "basic"),
        dropouts: Tuple[float, ...] = (0.0, 0.0),
        biases: Tuple[bool, ...] = (False, False),
        layer_scales: Tuple[bool, ...] = (False, False),
        activation: str = "star_relu",
        num_groups: int = 32,
        mlp_ratio: int = 2,
        slice_size: Optional[int] = 4,
        patch_embed_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        """Create a transformer for 2D-image-like (B, C, H, W) inputs.

        NOTE: The output shape is the same size as the input shape.

        Parameters
        ----------
            in_channels : int
                Number of channels in the 2D-input of shape (B, in_channels, H, W).
            num_heads : int, default=8
                Number of heads in multi-head attention.
            head_dim : int, default=64
                The out dim of the heads.
            cross_attention_dims : int, optional
                The out dim/length of the context query tensor. Cross attention combines
                asymmetrically two separate embeddings (context and input embeddings).
                E.g. passage from transformer encoder to transformer decoder. If this is
                set to None, no cross attention is applied.
            n_blocks : int, default=2
                Number of Multihead attention blocks in the transformer.
            block_types : Tuple[str, ...], default=("exact", "exact")
                The names/types of the SelfAttentionBlocks in the TransformerLayer.
                Length of the tuple has to equal `n_blocks`.
                Allowed names: ("exact", "linformer").
            computation_types : Tuple[str, ...], default=("basic", "basic")
                The way of computing the attention matrices in the SelfAttentionBlocks
                in the TransformerLayer. Length of the tuple has to equal `n_blocks`
                Allowed styles: "basic". "slice", "flash", "memeff", "slice_memeff".
            dropouts : Tuple[float, ...], default=(False, False)
                Dropout probabilities for the SelfAttention blocks.
            biases : bool, default=(True, True)
                Include bias terms in the SelfAttention blocks.
            layer_scales : bool, default=(False, False)
                Learnable layer weights for the self-attention matrix.
            activation : str, default="star_relu"
                The activation function applied at the end of the transformer layer fc.
                One of ("geglu", "approximate_gelu", "star_relu").
            num_groups : int, default=32
                Number of groups in the first group-norm op before the input is
                projected to be suitable for self-attention.
            mlp_ratio : int, default=2
                Scaling factor for the number of input features to get the number of
                hidden features in the final `Mlp` layer of the transformer.
            slice_size : int, optional, default=4
                Slice size for sliced self-attention. This is used only if
                `name = "slice"` for a SelfAttentionBlock.
            patch_embed_kwargs: Dict[str, Any], optional
                Extra key-word arguments for the patch embedding module. See the
                `ContiguousEmbed` module for more info.
        """
        super().__init__()
        patch_norm = "gn" if in_channels >= 32 else None
        self.patch_embed = ContiguousEmbed(
            in_channels=in_channels,
            patch_size=1,
            head_dim=head_dim,
            num_heads=num_heads,
            normalization=patch_norm,
            norm_kwargs={"num_features": in_channels, "num_groups": num_groups},
            **patch_embed_kwargs if patch_embed_kwargs is not None else {},
        )
        self.proj_dim = self.patch_embed.proj_dim

        self.transformer = TransformerLayer(
            query_dim=self.proj_dim,
            num_heads=num_heads,
            head_dim=head_dim,
            cross_attention_dim=cross_attention_dim,
            n_blocks=n_blocks,
            block_types=block_types,
            computation_types=computation_types,
            dropouts=dropouts,
            biases=biases,
            layer_scales=layer_scales,
            activation=activation,
            slice_size=slice_size,
            mlp_ratio=mlp_ratio,
            **kwargs,
        )

        self.proj_out = nn.Conv2d(
            self.proj_dim, in_channels, kernel_size=1, stride=1, padding=0
        )

    def forward(self, x: torch.Tensor, context: torch.Tensor = None) -> torch.Tensor:
        """Forward pass of the 2D transformer.

        Parameters
        ----------
            x : torch.Tensor
                Input image-like tensor. Shape (B, C, H, W).
            context : torch.Tensor, optional
                Context tensor. Shape (B, H*W, query_dim)). Can be used to add context
                information from another source than the current transformer.

        Returns
        -------
            torch.Tensor:
                Self-attended input tensor. Same shape as output.
        """
        B, _, H, W = x.shape
        residual = x

        # 1. embed and project
        x = self.patch_embed(x)

        # 2. transformer
        x = self.transformer(x, context)

        # 3. Reshape back to image-like shape.
        p_H = self.patch_embed.get_patch_size(H)
        p_W = self.patch_embed.get_patch_size(W)
        x = x.reshape(B, p_H, p_W, self.proj_dim).permute(0, 3, 1, 2)

        # Upsample to input dims if patch size less than orig inp size
        # assumes that the input is square mat.
        # NOTE: the kernel_size, pad, & stride has to be set correctly for this to work
        if p_H < H:
            scale_factor = H // p_H
            x = F.interpolate(x, scale_factor=int(scale_factor), mode="bilinear")

        # 4. project to original input channels
        x = self.proj_out(x)

        # 5. residual
        return x + residual


class TransformerLayer(nn.Module):
    def __init__(
        self,
        query_dim: int,
        num_heads: int = 8,
        head_dim: int = 64,
        cross_attention_dim: int = None,
        activation: str = "star_relu",
        n_blocks: int = 2,
        block_types: Tuple[str, ...] = ("exact", "exact"),
        computation_types: Tuple[str, ...] = ("basic", "basic"),
        dropouts: Tuple[float, ...] = (0.0, 0.0),
        biases: Tuple[bool, ...] = (False, False),
        layer_scales: Tuple[bool, ...] = (False, False),
        mlp_ratio: int = 2,
        slice_size: Optional[int] = 4,
        **kwargs,
    ) -> None:
        """Chain transformer blocks to compose a full generic transformer.

        NOTE: For 2D image like data:
            - Forward input shape: (B, H*W, head_dim*num_heads)
            - Forward output sahpe: (B, H*W, head_dim*num_heads)

        Parameters
        ----------
            query_dim : int
                The length/dim of the query. Typically: num_heads*head_dim
            num_heads : int, default=8
                Number of heads in multi-head attention.
            head_dim : int, default=64
                The out dim of the heads.
            cross_attention_dims : int, optional
                The out dim/length of the context query tensor. Cross attention combines
                asymmetrically two separate embeddings (context and input embeddings).
                E.g. passage from transformer encoder to transformer decoder. If this is
                set to None, no cross attention is applied.
            activation : str, default="star_relu"
                The activation function applied at the end of the transformer layer fc.
                One of ("gelu", "geglu", "approximate_gelu", "star_relu").
            n_blocks : int, default=2
                Number of SelfAttentionBlocks used in this layer.
            block_types : Tuple[str, ...], default=("exact", "exact")
                The name/type of the SelfAttentionBlocks in the TransformerLayer.
                Length of the tuple has to equal `n_blocks`.
                Allowed names: ("exact", "linformer").
            computation_types : Tuple[str, ...], default=("basic", "basic")
                The way of computing the attention matrices in the SelfAttentionBlocks
                in the TransformerLayer. Length of the tuple has to equal `n_blocks`
                Allowed styles: "basic". "slice", "flash", "memeff", "slice_memeff".
            dropouts : Tuple[float, ...], default=(False, False)
                Dropout probabilities for the SelfAttention blocks.
            biases : bool, default=(True, True)
                Include bias terms in the SelfAttention blocks.
            layer_scales : bool, default=(False, False)
                Learnable layer weights for the self-attention matrix.
            mlp_ratio : int, default=2
                Scaling factor for the number of input features to get the number of
                hidden features in the final `Mlp` layer of the transformer.
            slice_size : int, optional, default=4
                Slice size for sliced self-attention. This is used only if
                `name = "slice"` for a SelfAttentionBlock.
            **kwargs:
                Arbitrary key-word arguments.

        Raises
        ------
            ValueError: If the lengths of the tuple args are not equal to `n_blocks`.
        """
        super().__init__()

        illegal_args = [
            (k, a)
            for k, a in locals().items()
            if isinstance(a, tuple) and len(a) != n_blocks
        ]

        if illegal_args:
            raise ValueError(
                f"All the tuple-arg lengths need to be equal to `n_blocks`={n_blocks}. "
                f"Illegal args: {illegal_args}"
            )

        # self.tr_blocks = nn.ModuleDict()
        self.tr_blocks = nn.ModuleList()
        self.layer_scales = nn.ModuleList()
        blocks = list(range(n_blocks))
        for i in blocks:
            cross_dim = cross_attention_dim if i == blocks[-1] else None

            att_block = SelfAttentionBlock(
                name=block_types[i],
                how=computation_types[i],
                query_dim=query_dim,
                num_heads=num_heads,
                head_dim=head_dim,
                cross_attention_dim=cross_dim,
                dropout=dropouts[i],
                biases=biases[i],
                slice_size=slice_size,
                **kwargs,
            )
            self.tr_blocks.append(att_block)

            # add layer scale. (Optional)
            ls = LayerScale(query_dim) if layer_scales[i] else Identity()
            self.layer_scales.append(ls)

        self.mlp = MlpBlock(
            in_channels=query_dim,
            mlp_ratio=mlp_ratio,
            activation=activation,
            normalization="ln",
            norm_kwargs={"normalized_shape": query_dim},
            act_kwargs=kwargs,
        )

    def forward(self, x: torch.Tensor, context: torch.Tensor = None) -> torch.Tensor:
        """Forward pass of the transformer layer.

        Parameters
        ----------
            x : torch.Tensor
                Input query. Shape (B, H*W, query_dim)).
            context : torch.Tensor, optional
                Context tensor. Shape (B, H*W, query_dim)). Can be used to add context
                information from another source than the current layer.

        Returns
        -------
            torch.Tensor:
                Self-attended input tensor. Shape (B, H*W, query_dim).
        """
        n_blocks = len(self.tr_blocks)

        for i, (tr_block, ls) in enumerate(zip(self.tr_blocks, self.layer_scales), 1):
            # apply context only at the last transformer block
            con = None
            if i == n_blocks:
                con = context

            x = tr_block(x, con)
            x = ls(x)

        return self.mlp(x) + x



import pytest
import torch

from cellseg_models_pytorch.modules import Transformer2D


@pytest.mark.parametrize("block_type", ["exact", "linformer"])
@pytest.mark.parametrize("computation_type", ["basic", "slice"])
def test_transformer(block_type, computation_type):
    in_channels = 64
    B = 4
    H = W = 32

    x = torch.rand([B, in_channels, H, W])
    tr = Transformer2D(
        in_channels=in_channels,
        num_heads=4,
        head_dim=32,
        n_blocks=1,
        block_types=(block_type,),
        computation_types=(computation_type,),
        biases=(False,),
        dropouts=(0.0,),
        layer_scales=(False,),
        slice_size=4,
        seq_len=H * W,
    )

    out = tr(x)

    assert out.shape == x.shape



"""Copied the polygons to label utilities from stardist repo (with minor modifications).

BSD 3-Clause License

Copyright (c) 2018-2022, Uwe Schmidt, Martin Weigert
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from typing import Tuple

import numpy as np
from skimage.draw import polygon
from skimage.morphology import disk, erosion

from .nms import get_bboxes, nms_stardist

__all__ = ["post_proc_stardist_orig", "polygons_to_label"]


def polygons_to_label_coord(
    coord: np.ndarray, shape: Tuple[int, int], labels: np.ndarray = None
) -> np.ndarray:
    """Render polygons to image given a shape.

    Parameters
    ----------
        coord.shape : np.ndarray
            Shape: (n_polys, n_rays)
        shape : Tuple[int, int]
            Shape of the output mask.
        labels : np.ndarray, optional
            Sorted indices of the centroids.

    Returns
    -------
        np.ndarray:
            Instance labelled mask. Shape: (H, W).
    """
    coord = np.asarray(coord)
    if labels is None:
        labels = np.arange(len(coord))

    assert coord.ndim == 3 and coord.shape[1] == 2 and len(coord) == len(labels)

    lbl = np.zeros(shape, np.int32)

    for i, c in zip(labels, coord):
        rr, cc = polygon(*c, shape)
        lbl[rr, cc] = i + 1

    return lbl


def ray_angles(n_rays: int = 32):
    """Get linearly spaced angles for rays."""
    return np.linspace(0, 2 * np.pi, n_rays, endpoint=False)


def dist_to_coord(
    dist: np.ndarray, points: np.ndarray, scale_dist: Tuple[int, int] = (1, 1)
) -> np.ndarray:
    """Convert list of distances and centroids from polar to cartesian coordinates.

    Parameters
    ----------
        dist : np.ndarray
            The centerpoint pixels of the radial distance map. Shape (n_polys, n_rays).
        points : np.ndarray
            The centroids of the instances. Shape: (n_polys, 2).
        scale_dist : Tuple[int, int], default=(1, 1)
            Scaling factor.

    Returns
    -------
        np.ndarray:
            Cartesian cooridnates of the polygons. Shape (n_polys, 2, n_rays).
    """
    dist = np.asarray(dist)
    points = np.asarray(points)
    assert (
        dist.ndim == 2
        and points.ndim == 2
        and len(dist) == len(points)
        and points.shape[1] == 2
        and len(scale_dist) == 2
    )
    n_rays = dist.shape[1]
    phis = ray_angles(n_rays)
    coord = (dist[:, np.newaxis] * np.array([np.sin(phis), np.cos(phis)])).astype(
        np.float32
    )
    coord *= np.asarray(scale_dist).reshape(1, 2, 1)
    coord += points[..., np.newaxis]
    return coord


def _ind_prob_thresh(prob: np.ndarray, prob_thresh: float, b: int = 2) -> np.ndarray:
    """Index based thresholding."""
    if b is not None and np.isscalar(b):
        b = ((b, b),) * prob.ndim

    ind_thresh = prob > prob_thresh
    if b is not None:
        _ind_thresh = np.zeros_like(ind_thresh)
        ss = tuple(
            slice(_bs[0] if _bs[0] > 0 else None, -_bs[1] if _bs[1] > 0 else None)
            for _bs in b
        )
        _ind_thresh[ss] = True
        ind_thresh &= _ind_thresh
    return ind_thresh.astype("int32")


def polygons_to_label(
    dist: np.ndarray,
    points: np.ndarray,
    shape: Tuple[int, int],
    prob: np.ndarray = None,
    thresh: float = -np.inf,
    scale_dist: Tuple[int, int] = (1, 1),
) -> np.ndarray:
    """Convert distances and center points to instance labelled mask.

    Parameters
    ----------
        dist : np.ndarray
            The centerpoint pixels of the radial distance map. Shape (n_polys, n_rays).
        points : np.ndarray
            The centroids of the instances. Shape: (n_polys, 2).
        shape : Tuple[int, int]:
            Shape of the output mask.
        prob : np.ndarray, optional
            The centerpoint pixels of the regressed distance transform.
            Shape: (n_polys, n_rays).
        thresh : float, default=-np.inf
            Threshold for the regressed distance transform.
        scale_dist : Tuple[int, int], default=(1, 1)
            Scaling factor.

    Returns
    -------
        np.ndarray:
            Instance labelled mask. Shape (H, W).
    """
    dist = np.asarray(dist)
    points = np.asarray(points)
    prob = np.inf * np.ones(len(points)) if prob is None else np.asarray(prob)

    assert dist.ndim == 2 and points.ndim == 2 and len(dist) == len(points)
    assert len(points) == len(prob) and points.shape[1] == 2 and prob.ndim == 1

    ind = prob > thresh
    points = points[ind]
    dist = dist[ind]
    prob = prob[ind]

    ind = np.argsort(prob, kind="stable")
    points = points[ind]
    dist = dist[ind]

    coord = dist_to_coord(dist, points, scale_dist=scale_dist)

    return polygons_to_label_coord(coord, shape=shape, labels=ind)


def post_proc_stardist(
    dist_map: np.ndarray,
    stardist_map: np.ndarray,
    score_thresh: float = 0.5,
    iou_thresh: float = 0.5,
    trim_bboxes: bool = True,
    **kwargs,
) -> np.ndarray:
    """Run post-processing for stardist outputs.

    NOTE: This is not the original cpp version.
    This is a python re-implementation of the stardidst post-processing
    pipeline that uses non-maximum-suppression. Here, critical parts of the
    nms are accelerated with `numba` and `scipy.spatial.KDtree`.

    NOTE:
    This implementaiton of the stardist post-processing is actually nearly 2x
    faster than the original version if `trim_bboxes` is set to True. The resulting
    segmentation is not an exact match but the differences are mostly neglible.

    Parameters
    ----------
        dist_map : np.ndarray
            Predicted distance transform. Shape: (H, W).
        stardist_map : np.ndarray
            Predicted radial distances. Shape: (n_rays, H, W).
        thresh : float, default=0.4
            Threshold for the regressed distance transform.
        trim_bboxes : bool, default=True
            If True, The non-zero pixels are computed only from the cell contours
            which prunes down the pixel search space drastically.

    Returns
    -------
        np.ndarray:
            Instance labelled mask. Shape: (H, W).
    """
    if (
        not dist_map.ndim == 2
        and not stardist_map.ndim == 3
        and not dist_map.shape == stardist_map.shape[:2]
    ):
        raise ValueError(
            "Illegal input shapes. Make sure that: "
            f"`dist_map` has to have shape: (H, W). Got: {dist_map.shape} "
            f"`stardist_map` has to have shape (H, W, nrays). Got: {stardist_map.shape}"
        )

    dist = np.asarray(stardist_map).transpose(1, 2, 0)
    prob = np.asarray(dist_map)

    # threshold the edt distance transform map
    mask = _ind_prob_thresh(prob, score_thresh)

    # get only the mask contours to trim down bbox search space
    if trim_bboxes:
        fp = disk(2)
        mask -= erosion(mask, fp)

    points = np.stack(np.where(mask), axis=1)

    # Get only non-zero pixels of the transforms
    dist = dist[mask > 0]
    scores = prob[mask > 0]

    # sort descendingly
    ind = np.argsort(scores)[::-1]
    dist = dist[ind]
    scores = scores[ind]
    points = points[ind]

    # get bounding boxes
    x1, y1, x2, y2, areas, max_dist = get_bboxes(dist, points)
    boxes = np.stack([x1, y1, x2, y2], axis=1)

    # consider only boxes above score threshold
    score_cond = scores >= score_thresh
    boxes = boxes[score_cond]
    scores = scores[score_cond]
    areas = areas[score_cond]

    # run nms
    inds = nms_stardist(
        boxes,
        points,
        scores,
        areas,
        max_dist,
        score_threshold=score_thresh,
        iou_threshold=iou_thresh,
    )

    # get the centroids
    points = points[inds]
    scores = scores[inds]
    dist = dist[inds]
    labels = polygons_to_label(dist, points, prob=scores, shape=dist_map.shape)

    return labels


def post_proc_stardist_orig(
    dist_map: np.ndarray, stardist_map: np.ndarray, thresh: float = 0.4, **kwargs
) -> np.ndarray:
    """Run the original stardist post-processing pipeline.

    NOTE: to use this the `stardist` package needs to be installed.

    Parameters
    ----------
        dist_map : np.ndarray
            Predicted distance transform. Shape: (H, W).
        stardist_map : np.ndarray
            Predicted radial distances. Shape: (n_rays, H, W).
        thresh : float, default=0.4
            Threshold for the regressed distance transform.

    Returns
    -------
        np.ndarray:
            Instance labelled mask. Shape: (H, W).
    """
    try:
        from stardist import non_maximum_suppression
    except ImportError:
        raise ImportError(
            "Install stardist if using the original post-proc method. "
            "`pip install stardist`"
        )
    rescale = (1, 1)
    img_shape = dist_map.shape
    stardist_map = stardist_map.transpose(1, 2, 0)
    points, probi, disti = non_maximum_suppression(
        stardist_map, dist_map, prob_thresh=thresh
    )
    labels = polygons_to_label(
        disti, points, prob=probi, shape=img_shape, scale_dist=rescale
    )

    return labels



import torch
import torch.nn as nn

__all__ = ["BaseSelfAttention"]


class BaseSelfAttention(nn.Module):
    def __init__(
        self,
        head_dim: int,
        num_heads: int,
        how: str = "basic",
        slice_size: int = None,
        **kwargs,
    ) -> None:
        """Initialize a base class for self-attention modules.

        Four variants:
        - basic: self-attention implementation with torch.matmul O(N^2)
        - slice-attention: Computes the attention matrix in slices to save mem.
        - memeff: `xformers.ops.memory_efficient_attention` from xformers package.
        - slice-memeff-attention: Comnbines slice-attention and memeff

        Parameters
        ----------
            head_dim : int
                Out dim per attention head.
            num_heads : int
                Number of heads.
            how : str, default="basic"
                How to compute the self-attention matrix.
                One of ("basic", "flash", "slice", "memeff", "slice-memeff").
                "basic": the normal O(N^2) self attention.
                "flash": the flash attention (by xformers library),
                "slice": batch sliced attention operation to save mem.
                "memeff": xformers.memory_efficient_attention.
                "slice-memeff": Conmbine slicing and memory_efficient_attention.
            slice_size, int, optional
                The size of the slice. Used only if `how in ('slice', 'slice_memeff)`.

        Raises
        ------
            - ValueError:
                - If illegal self attention (`how`) method is given.
                - If `how` is set to `slice` while `num_heads` | `slice_size`
                    args are not given proper integer values.
                - If `how` is set to `memeff` or `slice_memeff` but cuda is not
                    available.
            - ModuleNotFoundError:
                - If `self_attention` is set to `memeff` and `xformers` package is not
                installed
        """
        super().__init__()

        allowed = ("basic", "flash", "slice", "memeff", "slice-memeff")
        if how not in allowed:
            raise ValueError(
                f"Illegal exact self attention type given. Got: {how}. "
                f"Allowed: {allowed}."
            )

        self.how = how
        self.head_dim = head_dim
        self.num_heads = num_heads

        if how == "slice":
            if any(s is None for s in (slice_size, num_heads)):
                raise ValueError(
                    "If `how` is set to 'slice', `slice_size`, `num_heads`, "
                    f"need to be given integer values. Got: `slice_size`: {slice_size} "
                    f"and `num_heads`: {num_heads}."
                )

        if how in ("memeff", "slice-memeff"):
            try:
                import xformers  # noqa F401
            except ModuleNotFoundError:
                raise ModuleNotFoundError(
                    "`self_attention` was set to `memeff`. The method requires the "
                    "xformers package. See how to install xformers: "
                    "https://github.com/facebookresearch/xformers"
                )
            if not torch.cuda.is_available():
                raise ValueError(
                    f"`how` was set to {how}. This method for computing self attentiton"
                    " is implemented  with `xformers.memory_efficient_attention` that "
                    "requires cuda."
                )

        # for slice_size > 0 the attention score computation
        # is split across the batch axis to save memory
        self.slice_size = slice_size
        self.proj_channels = self.head_dim * self.num_heads
